from phoebe.parameters import *
from phoebe.parameters.unit_choices import unit_choices as _unit_choices
import numpy as np

from .common import _use_autofig, _mplcmaps, _mplcolors, _mplmarkers, _mpllinestyles, MPLPropCycler, _label_units_lims, _figure_style_sources, _figure_style_nosources, _figure_uncover_highlight_animate


def lc(b, **kwargs):
    params = []

    # TODO: set hierarchy needs to update choices of any x,y,z in context='figure' to include phases:* a la pblum_ref

    params += [SelectParameter(qualifier='contexts', value=kwargs.get('contexts', '*'), choices=['dataset', 'model'], description='Contexts to include in the plot')]
    params += [SelectParameter(qualifier='datasets', value=kwargs.get('datasets', '*'), choices=[''], description='Datasets to include in the plot')]
    params += [SelectParameter(qualifier='models', value=kwargs.get('models', '*'), choices=[''], description='Models to include in the plot')]

    params += [ChoiceParameter(qualifier='x', value=kwargs.get('x', 'times'), choices=['times', 'phases'], description='Array to plot along x-axis')]
    params += [ChoiceParameter(qualifier='period', visible_if='[context][figure][kind]dperdt:!0.0,x:phases', value=kwargs.get('period', 'period'), choices=['period', 'period_anom'], advanced=True, description='Period to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='dpdt', visible_if='[context][figure][kind]dpdt:!0.0,x:phases', value=kwargs.get('dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='t0', visible_if='x:phases', value=kwargs.get('t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='y', value=kwargs.get('y', 'fluxes'), choices=['fluxes', 'residuals'], description='Array to plot along y-axis')]

    params += _label_units_lims('x', visible_if='x:times', default_unit=u.d, is_default=True, **kwargs)
    params += _label_units_lims('x', visible_if='x:!times', default_unit=u.dimensionless_unscaled, is_default=False, **kwargs)

    params += _label_units_lims('y', default_unit=u.W/u.m**2, is_default=True, **kwargs)

    kwargs.setdefault('color', 'black' if _use_autofig else None)
    kwargs.setdefault('marker', '.' if _use_autofig else None)
    kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    params += _figure_style_sources(b, default_color='model', default_marker='manual', default_linestyle='manual', **kwargs)

    params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    params += _figure_uncover_highlight_animate(b, **kwargs)

    return ParameterSet(params)


def rv(b, **kwargs):
    params = []

    params += [SelectParameter(qualifier='contexts', value=kwargs.get('contexts', '*'), choices=['dataset', 'model'], description='Contexts to include in the plot')]
    params += [SelectParameter(qualifier='datasets', value=kwargs.get('datasets', '*'), choices=[''], description='Datasets to include in the plot')]
    params += [SelectParameter(qualifier='models', value=kwargs.get('models', '*'), choices=[''], description='Models to include in the plot')]
    params += [SelectParameter(qualifier='components', value=kwargs.get('components', '*'), choices=[''], description='Components to include in the plot')]

    params += [ChoiceParameter(qualifier='x', value=kwargs.get('x', 'times'), choices=['times', 'phases'], description='Array to plot along x-axis')]
    params += [ChoiceParameter(qualifier='period', visible_if='[context][figure][kind]dperdt:!0.0,x:phases', value=kwargs.get('period', 'period'), choices=['period', 'period_anom'], advanced=True, description='Period to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='dpdt', visible_if='[context][figure][kind]dpdt:!0.0,x:phases', value=kwargs.get('dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='t0', visible_if='x:phases', value=kwargs.get('t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='y', value=kwargs.get('y', 'rvs'), choices=['rvs', 'residuals'], description='Array to plot along y-axis')]

    params += _label_units_lims('x', visible_if='x:times', default_unit=u.d, is_default=True, **kwargs)
    params += _label_units_lims('x', visible_if='x:!times', default_unit=u.dimensionless_unscaled, is_default=False, **kwargs)

    params += _label_units_lims('y', default_unit=u.km/u.s, is_default=True, **kwargs)

    kwargs.setdefault('color', 'black') if _use_autofig else None
    kwargs.setdefault('marker', '.' if _use_autofig else None)
    kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    params += _figure_style_sources(b, default_color='component', default_marker='manual', default_linestyle='manual', **kwargs)

    params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    params += _figure_uncover_highlight_animate(b, **kwargs)

    return ParameterSet(params)


# def etv(b, **kwargs):
#     params = []
#
#     params += [SelectParameter(qualifier='dataset', value=kwargs.get('dataset', '*'), choices=[''], description='Datasets to include in the plot')]
#     params += [SelectParameter(qualifier='model', value=kwargs.get('model', '*'), choices=[''], description='Models to include in the plot')]
#
#     params += [ChoiceParameter(qualifier='x', value=kwargs.get('x', 'time_ephems'), choices=['time_ephems', 'Ns'], description='Array to plot along x-axis')]
#     params += [ChoiceParameter(qualifier='y', value=kwargs.get('y', 'etvs'), choices=['etvs', 'time_ecls'], description='Array to plot along y-axis')]
#
#     params += _label_units_lims('x', visible_if='x:time_ephems', default_unit=u.d, is_default=True, **kwargs)
#     params += _label_units_lims('x', visible_if='x:Ns', default_unit=u.dimensionless_unscaled, **kwargs)
#
#     params += _label_units_lims('y', visible_if='y:etvs', default_unit=u.d, is_default=True, **kwargs)
#     params += _label_units_lims('y', visible_if='y:time_ecls', default_unit=u.d, **kwargs)
#
#     return ParameterSet(params)


def orb(b, **kwargs):
    params = []

    params += [SelectParameter(qualifier='datasets', value=kwargs.get('datasets', '*'), choices=[''], description='Datasets to include in the plot')]
    params += [SelectParameter(qualifier='models', value=kwargs.get('models', '*'), choices=[''], description='Models to include in the plot')]
    params += [SelectParameter(qualifier='components', value=kwargs.get('components', '*'), choices=[''], description='Components to include in the plot')]

    params += [ChoiceParameter(qualifier='x', value=kwargs.get('x', 'us'), choices=['times', 'phases', 'us', 'vs', 'ws', 'vus', 'vvs', 'vws'], description='Array to plot along x-axis')]
    params += [ChoiceParameter(qualifier='period', visible_if='[context][figure][kind]dperdt:!0.0,x:phases', value=kwargs.get('period', 'period'), choices=['period', 'period_anom'], advanced=True, description='Period to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='dpdt', visible_if='[context][figure][kind]dpdt:!0.0,x:phases', value=kwargs.get('dpdt', 'dpdt'), choices=['dpdt', 'none'], advanced=True, description='dpdt to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='t0', visible_if='x:phases', value=kwargs.get('t0', 't0_supconj'), choices=['t0_supconj', 't0_perpass', 't0_ref'], advanced=True, description='t0 to use when phasing for x-axis')]
    params += [ChoiceParameter(qualifier='y', value=kwargs.get('y', 'ws'), choices=['us', 'vs', 'ws', 'vus', 'vvs', 'vws'], description='Array to plot along y-axis')]

    params += _label_units_lims('x', visible_if='x:times', default_unit=u.d, is_default=False, **kwargs)
    # TODO: this visible_if will likely fail if/once we implement phases:* for the choices for x
    params += _label_units_lims('x', visible_if='x:phases', default_unit=u.dimensionless_unscaled, is_default=False, **kwargs)
    params += _label_units_lims('x', visible_if='x:us|vs|ws', default_unit=u.solRad, is_default=True, **kwargs)
    params += _label_units_lims('x', visible_if='x:vus|vvs|vws', default_unit=u.km/u.s, is_default=False, **kwargs)

    params += _label_units_lims('y', visible_if='y:us|vs|ws', default_unit=u.solRad, is_default=True, **kwargs)
    params += _label_units_lims('y', visible_if='y:vus|vvs|vws', default_unit=u.km/u.s, is_default=False, **kwargs)

    kwargs.setdefault('color', 'black' if _use_autofig else None)
    kwargs.setdefault('marker', '.' if _use_autofig else None)
    kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    params += _figure_style_sources(b, default_color='component', default_marker='manual', default_linestyle='manual', **kwargs)

    params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    params += _figure_uncover_highlight_animate(b, **kwargs)

    return ParameterSet(params)

def lp(b, **kwargs):
    params = []

    params += [SelectParameter(qualifier='contexts', value=kwargs.get('contexts', '*'), choices=['dataset', 'model'], description='Contexts to include in the plot')]
    params += [SelectParameter(qualifier='datasets', value=kwargs.get('datasets', '*'), choices=[''], description='Datasets to include in the plot')]
    params += [SelectParameter(qualifier='models', value=kwargs.get('models', '*'), choices=[''], description='Models to include in the plot')]
    # params += [SelectParameter(qualifier='times', value=kwargs.get('times', '*'), choices=[''], description='Times to include in the plot')]
    params += [SelectParameter(qualifier='components', value=kwargs.get('components', '*'), choices=[''], description='Components to include in the plot')]

    params += [ChoiceParameter(qualifier='x', value=kwargs.get('x', 'wavelengths'), choices=['wavelengths'], description='Array to plot along x-axis')]
    params += [ChoiceParameter(qualifier='y', value=kwargs.get('y', 'flux_densities'), choices=['flux_densities'], description='Array to plot along y-axis')]

    params += _label_units_lims('x', default_unit=u.nm, is_default=True, **kwargs)
    params += _label_units_lims('y', default_unit=u.W/(u.m**2*u.nm), is_default=True, **kwargs)

    kwargs.setdefault('color', 'black' if _use_autofig else None)
    kwargs.setdefault('marker', '.' if _use_autofig else None)
    kwargs.setdefault('linestyle', 'solid' if _use_autofig else None)
    params += _figure_style_sources(b, default_color='component', default_marker='manual', default_linestyle='manual', **kwargs)

    params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', True), advanced=True, description='Whether to draw the legend')]

    params += _figure_uncover_highlight_animate(b, **kwargs)

    return ParameterSet(params)

def mesh(b, **kwargs):
    params = []

    params += [SelectParameter(qualifier='datasets', value=kwargs.get('datasets', '*'), choices=[''], description='Datasets to include in the plot')]
    params += [SelectParameter(qualifier='models', value=kwargs.get('models', '*'), choices=[''], description='Models to include in the plot')]
    # params += [SelectParameter(qualifier='times', value=kwargs.get('times', '*'), choices=[''], description='Times to include in the plot')]
    params += [SelectParameter(qualifier='components', value=kwargs.get('components', '*'), choices=[''], description='Components to include in the plot')]

    params += [ChoiceParameter(qualifier='x', value=kwargs.get('x', 'us'), choices=['us', 'vs', 'ws', 'xs', 'ys', 'zs'], description='Array to plot along x-axis')]
    params += [ChoiceParameter(qualifier='y', value=kwargs.get('y', 'vs'), choices=['us', 'vs', 'ws', 'xs', 'ys', 'zs'], description='Array to plot along y-axis')]


    params += _label_units_lims('x', visible_if='x:us|vs|ws', default_unit=u.solRad, is_default=True, **kwargs)
    params += _label_units_lims('x', visible_if='x:xs|ys|zs', default_unit=u.dimensionless_unscaled, is_default=False, **kwargs)

    params += _label_units_lims('y', visible_if='y:us|vs|ws', default_unit=u.solRad, is_default=True, **kwargs)
    params += _label_units_lims('y', visible_if='y:xs|ys|zs', default_unit=u.dimensionless_unscaled, is_default=False, **kwargs)

    params += [ChoiceParameter(qualifier='fc_source', value=kwargs.get('fc_source', 'column'), choices=['column', 'manual', 'component', 'model'], description='Source to use for facecolor.  For column, see the fc_column parameter.  For manual, see the fc parameter.  Otherwise, see the color parameter tagged with the corresponding component/model')]
    params += [ChoiceParameter(qualifier='fc_column', visible_if='fc_source:column', value=kwargs.get('fc_column', 'None'), choices=['None'], description='Column from the mesh to plot as facecolor if fc_source is column.')]
    params += [ChoiceParameter(qualifier='fc', visible_if='fc_source:manual', value=kwargs.get('fc', 'None'), choices=['None']+_mplcolors, description='Color to use as facecolor if fc_source is manual.')]
    params += [ChoiceParameter(qualifier='fcmap_source', visible_if='fc_source:column', value=kwargs.get('fcmap_source', 'auto'), choices=['auto', 'manual'], description='Source to use for facecolor colormap.  To have fcmap adjust based on the value of fc, use auto.  For manual, see the fcmap parameter.')]
    params += [ChoiceParameter(qualifier='fcmap', visible_if='fc_source:column,fcmap_source:manual', value=kwargs.get('fcmap', 'viridis'), choices=_mplcmaps, description='Colormap to use for facecolor if fcmap_source=\'manual\'.')]

    params += [ChoiceParameter(qualifier='ec_source', value=kwargs.get('ec_source', 'manual'), choices=['column', 'manual', 'component', 'model', 'face'], description='Source to use for edgecolor.  For column, see the ec_column parameter.  For manual, see the ec parameter.  Otherwise, see the color parameter tagged with the corresponding component/model')]
    params += [ChoiceParameter(qualifier='ec_column', visible_if='ec_source:column', value=kwargs.get('ec_column', 'None'), choices=['None'], description='Column from the mesh to plot as edgecolor if ec_source is column.')]
    params += [ChoiceParameter(qualifier='ec', visible_if='ec_source:manual', value=kwargs.get('ec', 'black' if _use_autofig else None), choices=['face']+_mplcolors, description='Color to use as edgecolor if ec_source is manual.')]
    params += [ChoiceParameter(qualifier='ecmap_source', visible_if='ec_source:column', value=kwargs.get('ecmap_source', 'auto'), choices=['auto', 'manual'], description='Source to use for edgecolor colormap.  To have ecmap adjust based on the value of ec, use auto.  For manual, see the ecmap parameter.')]
    params += [ChoiceParameter(qualifier='ecmap', visible_if='ec_source:column,ecmap_source:manual', value=kwargs.get('fcmap', 'viridis'), choices=_mplcmaps, description='Colormap to use for edgecolor if ecmap_source=\'manual\'.')]




    for q in ['fc', 'ec']:
        # see dataset._mesh_columns
        # even though this isn't really the default case, we'll pass is_default=True
        # so that we get fc/eclim_source, etc, parameters
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:vxs|vys|vzs|vus|vws|vvs|rvs'.format(q,q), default_unit=u.km/u.s, is_default=True, **kwargs)
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:rs|rprojs'.format(q,q), default_unit=u.solRad, is_default=False, **kwargs)
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:areas'.format(q,q), default_unit=u.solRad**2, is_default=False, **kwargs)
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:teffs'.format(q,q), default_unit=u.K, is_default=False, **kwargs)
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:pblum_ext|abs_pblum_ext'.format(q,q), default_unit=u.W, is_default=False, **kwargs)
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:ptfarea'.format(q,q), default_unit=u.m, is_default=False, **kwargs)
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:intensities|normal_intensities|abs_intensities|abs_normal_intensities'.format(q,q), default_unit=u.W/u.m**3, is_default=False, **kwargs)
        params += _label_units_lims(q, visible_if='{}_source:column,{}_column:visibilities|mus|loggs|boost_factors|ldint'.format(q,q), default_unit=u.dimensionless_unscaled, is_default=False, **kwargs)

    # TODO: legend=True currently fails
    params += [BoolParameter(qualifier='draw_sidebars', value=kwargs.get('draw_sidebars', True), advanced=True, description='Whether to draw the sidebars')]
    params += [BoolParameter(qualifier='legend', value=kwargs.get('legend', False), advanced=True, description='Whether to draw the legend')]

    params += _figure_uncover_highlight_animate(b, uncover=False, highlight=False, **kwargs)

    return ParameterSet(params)



"""
this is a python implementation for a forward propagation of a pre-trained EBAI
ANN (see https://github.com/aprsa/ebai)
"""

import numpy as np
import os

def _activate(val, theta=0.75, tau=0.35):
	return 1./(1.+np.exp(-(val-theta)/tau))

def _rescale(val, bounds):
    return bounds[:,0] + (val-0.1)/(0.9-0.1)*(bounds[:,1]-bounds[:,0])

_dir = os.path.dirname(__file__)
i2h = np.loadtxt(os.path.join(_dir, 'i2h.weights'))
h2o = np.loadtxt(os.path.join(_dir, 'h2o.weights'))

bounds = np.loadtxt(os.path.join(_dir, 'bounds.data'))

def ebai_forward(fluxes):
	if len(fluxes) != 201:
		raise ValueError("fluxes must have length of 201 (evenly sampled in phase-space)")

	return _rescale(_activate(np.matmul(h2o, _activate(np.matmul(i2h, fluxes)))), bounds=bounds)




import __main__
import re

class Completer:
    """
    [FUTURE]
    """
    def __init__(self, namespace = None):
        """Create a new completer for the command line.

        Completer([namespace]) -> completer instance.

        Completer instances should be used as the completion mechanism of
        readline via the set_completer() call:

        readline.set_completer(Completer(my_namespace).complete)
        """

        if namespace and not isinstance(namespace, dict):
            raise TypeError('namespace must be a dictionary')

        # Don't bind to namespace quite yet, but flag whether the user wants a
        # specific namespace or to use __main__.__dict__. This will allow us
        # to bind to __main__.__dict__ at completion time, not now.
        if namespace is None:
            self.use_main_ns = 1
        else:
            self.use_main_ns = 0
            self.namespace = namespace

    def complete(self, text, state):
        """Return the next possible completion for 'text'.

        This is called successively with state == 0, 1, 2, ... until it
        returns None.  The completion should begin with 'text'.

        """
        if self.use_main_ns:
            self.namespace = __main__.__dict__
        if state == 0:
            self.matches = self.attr_matches(text)
        try:
            return self.matches[state]
        except IndexError:
            return None

    def attr_matches(self, text):
        """Compute matches when text contains a dot.

        Assuming the text is of the form NAME.NAME....[NAME], and is
        evaluable in self.namespace, it will be evaluated and its attributes
        (as revealed by dir()) are used as possible completions.  (For class
        instances, class members are also considered.)

        WARNING: this can still invoke arbitrary C code, if an object
        with a __getattr__ hook is evaluated.

        """

        def _method_or_attr(thisobject, item):
            # decide whether to append a '(' to the end of the attr based
            # on whether its callable
            if hasattr(getattr(thisobject, item), '__call__'):
                return item + '('
            else:
                return item

        tb_compl_commands = {
                        '.': {},
                        '[': {},
                        '.get(': {},
                        '.set(': {},
                        '.filter(': {},
                        '.filter_or_get(': {},
                        '.get_parameter(': {},
                        '.remove_parameter(': {},
                        '.remove_parameters_all(': {},
                        '.get_value(': {},
                        '.set_value(': {},
                        '.set_value_all(': {},
                        # TODO: default_unit, adjust, prior, posterior, enabled?
                        '.get_history(': {'context': 'history'},
                        '.remove_history(': {'context': 'history'},
                        '.get_component(': {'context': 'system'},
                        '.remove_component(': {'context': 'system'},
                        '.get_mesh(': {'context': 'mesh'},
                        '.remove_mesh(': {'context': 'mesh'},
                        '.get_constraint(': {'context': 'constraint'},
                        '.remove_constraint(': {'context': 'constraint'},
                        '.flip_constraint(': {'context': 'constraint'},
                        '.run_constraint(': {'context': 'constraint'},
                        '.get_compute(': {'context': 'compute'},
                        '.remove_compute(': {'context': 'compute'},
                        '.run_compute(': {'context': 'compute'},
                        '.get_distribution(': {'context': 'distribution'},
                        '.sample_distribution(': {'context': 'distribution'},
                        '.remove_distribution(': {'context': 'distribution'},
                        '.get_solver(': {'context': 'solver'},
                        '.remove_solver(': {'context': 'solver'},
                        '.run_solver(': {'context': 'solver'},
                        '.get_solution(': {'context': 'solution'},
                        '.remove_solution(': {'context': 'solution'},
                        # TODO: plots, plugins
                        }

        expr = None
        for cmd,filter_kwargs in tb_compl_commands.items():
            if cmd in text:
                expr, attr = text.rsplit(cmd, 1)
                #~ if len(attr)==0:
                    #~ return []
                if attr[0] not in ["'",'"'] and cmd != '.':
                    return []
                else:
                    if cmd == '.':
                        # then we're just looking for attributes and don't
                        # need to offset for the ' or "
                        stringchar = ''
                        attr = attr
                    else:
                        # then we're the first argument of some method
                        # and need to account for the starting ' or "
                        stringchar = attr[0]
                        attr = attr[1:]
                    break

        if expr is None:
            # then we haven't found a match
            return []

        try:
            thisobject = eval(expr, self.namespace)
        except Exception:
            return []

        if cmd == '.':
            # then we're looking for attributes of thisobject (PS or bundle) that start with attr
            words = [_method_or_attr(thisobject, item) for item in dir(thisobject) if item[:len(attr)] == attr]
        else:
            # then we're looking to autocomplete the twig attr for thisobject (PS or bundle)
            words = thisobject.filter_or_get(attr, autocomplete=True, **filter_kwargs)

        matches = []
        n = len(attr)
        for word in words:
            matches.append('{}{}{}{}'.format(expr,cmd,stringchar,word))
        return matches



import numpy as np
from collections import OrderedDict
from copy import deepcopy as _deepcopy
import json
import sys

try:
    from astropy import units
except ImportError:
    _has_astropy = False
else:
    _has_astropy = True

if sys.version_info[0] == 3:
  unicode = str

################## VALIDATORS ###################

# these all must accept a single value and return a boolean if it matches the condition as well as any alterations to the value
# NOTE: the docstring is used as the error message if the test fails

def is_unit(value):
    """must be an astropy unit"""
    if not _has_astropy:
        raise ImportError("astropy must be installed for unit support")
    if (isinstance(value, units.Unit) or isinstance(value, units.IrreducibleUnit) or isinstance(value, units.CompositeUnit)):
        return True, value
    else:
        return False, value

def is_unit_or_unitstring(value):
    """must be an astropy.unit"""
    if is_unit(value)[0]:
        return True, value
    try:
        unit = units.Unit(value)
    except:
        return False, value
    else:
        return True, unit

def is_unit_or_unitstring_or_none(value):
    """must be an astropy unit or None"""
    if value is None:
        return True, value
    return is_unit_or_unitstring(value)

def is_bool(value):
    """must be boolean"""
    if isinstance(value, str) or isinstance(value, unicode):
        if value.upper() == 'TRUE':
            return True, True
        elif value.upper() == 'FALSE':
            return True, False
        else:
            return False, None

    return isinstance(value, bool), value

def is_float(value):
    """must be a float"""
    if isinstance(value, str) or isinstance(value, unicode):
        try:
            value = float(value)
        except:
            return False, None
        else:
            return True, value
    return isinstance(value, float) or isinstance(value, int) or isinstance(value, np.float64), float(value)

def is_int(value):
    """must be an integer"""
    if isinstance(value, str) or isinstance(value, unicode):
        if "." in value:
            return False, None

        try:
            value = int(float(value))
        except:
            return False, None
        else:
            return True, value
    return isinstance(value, int), value

def is_int_positive(value):
    """must be a positive integer"""
    _is_int, value = is_int(value)
    return _is_int and value > 0, value

def is_int_positive_or_none(value):
    """must be a postive integer or None"""
    if value is None:
        return True, value
    return is_int_positive(value)

def is_valid_shape(value):
    """must be a positive integer or a tuple/list of positive integers"""
    if is_int_positive(value):
        return True, value
    elif isinstance(value, tuple) or isinstance(value, list):
        for v in value:
            if not is_int_positive(v):
                return False, value
        return True, value
    else:
        return False, value

def is_iterable(value):
    """must be an iterable (list, array, tuple)"""
    return isinstance(value, np.ndarray) or isinstance(value, list) or isinstance(value, tuple), value


############# WRAPPERS ###################

class ArrayWrapper(object):
    """
    ArrayWrapper should be subclassed by any type of array creation classes
    and the corresponding function should be added to __init__.py.  Any subclass
    MUST define the following:

    __init__ (see docs below)
    array (@property - see docs below)
    __math__(self, operator, other)
    """
    def __init__(self, *args):
        """
        all subclasses MUST parse args and send in as tuples via super so that
        order is preserved.

        For example:
        def __init__(self, start, stop, step):
            super(MyClass, self).__init__(('start', start, is_float), ('stop', stop, is_float), ('step', step, is_float))

        All of these "descriptors" will then be available to get and set via
        their attribute names
        """
        self._descriptors = OrderedDict()
        self._validators = OrderedDict()

        for item in args:
            valid, validated_value = item[2](item[1])
            if valid:
                self._descriptors[item[0]] = validated_value
            else:
                raise ValueError("{} {}, got {}".format(item[0], item[2].__doc__, item[1]))
            self._validators[item[0]] = item[2]

    @property
    def array(self):
        """
        all subclasses MUST define a array property method which returns a
        np.ndarray object
        """
        raise NotImplementedError

    @property
    def quantity(self):
        """
        Return the underlying astropy quantity (if astropy is installed).

        Returns
        --------
        * astropy quantity
        """
        if not _has_astropy:
            raise ImportError("astropy must be installed for unit/quantity support")

        if self.unit is None:
            raise ValueError("unit is not set, cannot convert to quantity")

        return self.array * self.unit

    def to(self, unit):
        """
        Convert between units.  Returns a new nparray object with the new units.

        Arguments
        ---------
        * `unit` (astropy unit or string): new units

        Returns
        -------
        * nparray object of the same class as the original
        """
        if not _has_astropy:
            raise ImportError("astropy must be installed for unit/quantity support")

        if self.unit is None:
            raise ValueError("no units currently set")

        if not is_unit_or_unitstring(unit)[0]:
            raise ValueError("unit not recognized")

        mult_factor = self.unit.to(unit)
        copy = self.copy() * mult_factor
        copy.unit = unit
        return copy

    # @property
    # def __class__(self):
    #     """
    #     override the class so that isinstance(object, np.ndarray) returns True
    #
    #     NOTE: this is really hacky
    #     """
    #     return np.ndarray

    def __getattr__(self, name):
        """
        for anything that isn't overriden here, call the method on the array itself
        """
        if name in ['_descriptors', '_validators']:
            # then we need to actually get the attribute
            return super(ArrayWrapper, self).__getattr__(name)
        elif name in self._descriptors.keys():
            # then get the item in the dictionary
            return self._descriptors.get(name)
        elif hasattr(self.array, name):
            # then fallback on the underlying array object
            return getattr(self.array, name)
        else:
            raise AttributeError("neither '{}' or '{}' have attribute '{}'".format(self.__class__.__name__.lower(), 'numpy.ndarray', name))

    def __setattr__(self, name, value):
        """
        """
        if name in ['_descriptors', '_validators', '__class__']:
            return super(ArrayWrapper, self).__setattr__(name, value)
        elif name in self._descriptors.keys():
            valid, validated_value = self._validators[name](value)
            if valid:
                self._descriptors[name] = validated_value
            else:
                raise ValueError("{} {}".format(name, validator.__doc__))
        else:
            return setattr(self.array, name, value)

    def __getitem__(self, index):
        return self.array.__getitem__(index)

    def __setitem__(self, index, value):
        """
        setting an item in the array needs to fallback on the Array class
        """
        arrayvalue = self.array
        arrayvalue.__setitem__(index, value)
        self._convert_to_array(arrayvalue)


    # def __dir__(self):
    #     """
    #     override the dir() functionality to also include all of the
    #     methods/attributes of self.array
    #     """

    def __repr__(self):
        descriptors = " ".join(["{}={}".format(k,v) for k,v in self._descriptors.items()])
        return "<{} {}>".format(self.__class__.__name__.lower(), descriptors)

    def __copy__(self):
        return self.__class__(**self._descriptors)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = self.__class__.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, _deepcopy(v, memo))
        return result

    def copy(self):
        """
        Create and return a copy.
        """
        return self.__copy__()

    def to_dict(self):
        """
        Dump a representation of the nparray object to a dictionary.  The
        nparray object should then be able to be fully restored via
        <nparray.from_dict>

        Returns
        ----------
        * (dictionary)
        """
        def _json_safe(v):
            if isinstance(v, np.ndarray):
                return v.tolist()
            elif is_unit(v)[0]:
                return v.to_string()
            else:
                return v
        d = {k:_json_safe(v) for k,v in self._descriptors.items()}
        d['nparray'] = self.__class__.__name__.lower()
        return d

    def to_json(self, **kwargs):
        """
        Dump a representation of the nparray object to a json-formatted string.
        The nparray object should then be able to be fully restored via
        <nparray.from_json>.

        Arguments
        -----------
        * `**kwargs`: all keyword arguments are sent to json.dumps

        Returns
        ---------
        * (string): json formatted string.
        """
        return json.dumps(self.to_dict(), **kwargs)

    def to_file(self, filename, **kwargs):
        """
        Dump a representation of the nparray object to a json-formatted file.
        The nparray object should then be able to be fully restored via
        <nparray.from_file>.

        Arguments
        -----------
        * `filename` (string): path to the file to be created (will overwrite
            if already exists)
        * `**kwargs`: additional keyword arguments are passed to
            json.dumps.

        Returns
        -----------
        * (string): the filename
        """
        f = open(filename, 'w')
        f.write(self.to_json(**kwargs))
        f.close()
        return filename

    def to_array(self):
        """
        Convert to an nparray <Array> object (**note**: not a numpy array, to access
        the underlying numpy object, call the .array property)

        Returns
        ---------
        * <Array>
        """
        return Array(self.array)

    def _convert_to_array(self, value=None):
        if value is None:
            value = self.array
        self.__class__ = Array
        self.__init__(value)

    def __add__(self, other):
        return self.__math__('__add__', other)

    def __radd__(self, other):
        return self.__math__('__radd__', other)

    def __sub__(self, other):
        return self.__math__('__sub__', other)

    def __rsub__(self, other):
        return self.__math__('__rsub__', other)

    def __mul__(self, other):
        if _has_astropy and is_unit(other)[0]:
            copy = self.copy()
            copy.unit = other
            return copy
        else:
            return self.__math__('__mul__', other)

    def __rmul__(self, other):
        return self.__math__('__rmul__', other)

    def __div__(self, other):
        return self.__math__('__div__', other)

    def __rdiv__(self, other):
        return self.__math__('__rdiv__', other)

    def __pow__(self, other):
        return self.__math__('__pow__', other)

    def __rpow__(self, other):
        return Array(self.array.__rpow__(other))

    def __abs__(self):
        return Array(abs(self.array))

    def __len__(self):
        return len(self.array)

    @property
    def shape(self):
        if 'shape' in self._descriptors.keys():
            # Ones, Full, Zeros has a shape attribute directly
            shape = self._descriptors.get('shape')
            if isinstance(shape, int):
                return (shape,)
            return shape
        return self.array.shape

    def __comparison__(self, operator, other):
        """
        determine comparisons based on the underyling arrays
        """
        return getattr(self.array, operator)(other)

    def __eq__(self, other):
        return self.__comparison__('__eq__', other)

    def __ne__(self, other):
        return self.__comparison__('__ne__', other)

    def __lt__(self, other):
        return self.__comparison__('__lt__', other)

    def __lte__(self, other):
        return self.__comparison__('__le__', other)

    def __gt__(self, other):
        return self.__comparison__('__gt__', other)

    def __ge__(self, other):
        return self.__comparison__('__ge__', other)

    def __contains__(self, other):
        return self.__comparison__('__contains__', other)

class Array(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.array>.
    """
    def __init__(self, value, unit=None):
        """
        This is available as a top-level convenience function as <nparray.array>.

        Arguments
        ------------
        * `value` (array or list): array or list of values.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Array>
        """
        super(Array, self).__init__(('value', value, is_iterable), ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.array(value)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.array(self.value)

    def __math__(self, operator, other):
        value = getattr(np.asarray(self.value), operator)(other)
        return Array(value, self.unit)

    def __setitem__(self, index, value):
        """
        for Arrays, we simply edit the underlying numpy array/list/tuple
        """
        # TODO: will this cause issues if value is a tuple? what is the expected behavior
        self.value.__setitem__(index, value)

class Arange(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.arange>.
    """
    def __init__(self, start, stop, step, unit=None):
        """
        This is available as a top-level convenience function as <nparray.arange>.

        Arguments
        ------------
        * `start` (int or float): the starting point of the sequence.
        * `stop` (int or float): the ending point of the sequence.  The interval
            does not include this value, except in some cases where `step` is not an
            integer and floating point round-off affects the length of the array.
        * `step` (int or float): the stepsize between each item in the sequence.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Arange>
        """
        super(Arange, self).__init__(('start', start, is_float),
                                     ('stop', stop, is_float),
                                     ('step', step, is_float),
                                     ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.arange(start, stop, step)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.arange(self.start, self.stop, self.step)

    def to_linspace(self):
        """
        Convert from <Arange> to <Linspace>.

        ```py
        num = int((self.stop - self.start)/self.step)
        Linspace(self.start, self.stop-self.step, num)
        ```

        Returns
        -------
        * <Linspace>
        """
        num = int((self.stop-self.start)/self.step)
        return Linspace(self.start, self.stop-self.step, num)

    def to_array(self):
        """
        Convert from <Arange> to <Array> by executing <Arange.array>.

        Returns
        --------
        * <Array>
        """
        return Array(self.array)

    def __math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            # NOTE: we have to cast the start and stop to a float because
            # int.__mul__(float) is not implemented in Python (instead it uses
            # float.__rmul__(int)).
            # Alternatively,we could check for this case and do
            # other.__r{operator}(self.start) but that seems just as ugly.
            return Arange(getattr(float(self.start), operator)(other), getattr(float(self.stop), operator)(other), self.step, self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))

class Linspace(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.linspace>.
    """
    def __init__(self, start, stop, num, endpoint=True, unit=None):
        """
        This is available as a top-level convenience function as <nparray.linspace>.

        Arguments
        ------------
        * `start` (int or float): the starting point of the sequence.
        * `stop` (int or float): the ending point of the sequence, unless `endpoint`
            is set to False.  In that case, the sequence consists of all but the
            last of ``num + 1`` evenly spaced samples, so that `stop` is excluded.
            Note that the step size changes when `endpoint` is False.
        * `num` (int): number of samples to generate.
        * `endpoint` (bool, optional, default=True): If True, `stop` is the last
            sample. Otherwise, it is not included.
        * `unit` (astropy unit or string, optional, default=None): unit
            corresponding to the passed values.

        Returns
        -----------
        * <Linspace>
        """
        super(Linspace, self).__init__(('start', start, is_float),
                                       ('stop', stop, is_float),
                                       ('num', num, is_int_positive),
                                       ('endpoint', endpoint, is_bool),
                                       ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.linspace(start, stop, num, endpoint)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.linspace(self.start, self.stop, self.num, self.endpoint)

    def to_arange(self):
        """
        Convert from <Linspace> to <Arange>

        ```py
        arr, step = np.linspace(self.start, self.stop, self.num, self.endpoint, retstep=True)
        Arange(self.start, self.stop+step, step)
        ```

        Returns
        ----------
        * <Arange>
        """
        arr, step = np.linspace(self.start, self.stop, self.num, self.endpoint, retstep=True)
        return Arange(self.start, self.stop+step, step)

    def __math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            # NOTE: we have to cast the start and stop to a float because
            # int.__mul__(float) is not implemented in Python (instead it uses
            # float.__rmul__(int)).
            # Alternatively,we could check for this case and do
            # other.__r{operator}(self.start) but that seems just as ugly.
            return Linspace(getattr(float(self.start), operator)(other), getattr(float(self.stop), operator)(other), self.num, self.endpoint, self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))

class Invspace(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.invspace>.
    """
    def __init__(self, start, stop, num, endpoint=True, unit=None):
        """
        This is available as a top-level convenience function as <nparray.invspace>.

        Arguments
        ------------
        * `start` (int or float): the starting point of the sequence.
        * `stop` (int or float): the ending point of the sequence, unless `endpoint`
            is set to False.  In that case, the sequence consists of all but the
            last of ``num + 1`` evenly spaced samples, so that `stop` is excluded.
            Note that the step size changes when `endpoint` is False.
        * `num` (int): number of samples to generate.
        * `endpoint` (bool, optional, default=True): If True, `stop` is the last
            sample. Otherwise, it is not included.
        * `unit` (astropy unit or string, optional, default=None): unit
            corresponding to the passed values.

        Returns
        -----------
        * <Invspace>
        """
        super(Invspace, self).__init__(('start', start, is_float),
                                       ('stop', stop, is_float),
                                       ('num', num, is_int_positive),
                                       ('endpoint', endpoint, is_bool),
                                       ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        1./np.linspace(1./start, 1./stop, num, endpoint)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return 1./np.linspace(1./self.start, 1./self.stop, self.num, self.endpoint)

    def __math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            # NOTE: we have to cast the start and stop to a float because
            # int.__mul__(float) is not implemented in Python (instead it uses
            # float.__rmul__(int)).
            # Alternatively,we could check for this case and do
            # other.__r{operator}(self.start) but that seems just as ugly.
            return Invspace(getattr(float(self.start), operator)(other), getattr(float(self.stop), operator)(other), self.num, self.endpoint, self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))

class Logspace(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.logspace>.
    """
    def __init__(self, start, stop, num, endpoint=True, base=10.0, unit=None):
        """
        This is available as a top-level convenience function as <nparray.logspace>.

        Arguments
        ------------
        * `start` (int or float): ``base ** start`` is the starting value of the sequence.
        * `stop` (int or float): ``base ** stop`` is the final value of the sequence,
            unless `endpoint` is False.  In that case, ``num + 1`` values are spaced
            over the interval in log-space, of which all but the last (a sequence of
            length `num`) are returned.
        * `num` (int): number of samples to generate.
        * `endpoint` (bool, optional, default=True): If True, `stop` is the last
            sample. Otherwise, it is not included.
        * `base` (float, optional, default=10.0): The base of the log space. The
            step size between the elements in ``ln(samples) / ln(base)``
            (or ``log_base(samples)``) is uniform.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Logspace>
        """
        super(Logspace, self).__init__(('start', start, is_float),
                                       ('stop', stop, is_float),
                                       ('num', num, is_int_positive),
                                       ('endpoint', endpoint, is_bool),
                                       ('base', base, is_float),
                                       ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.logspace(start, stop, num, endpoint, base)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.logspace(self.start, self.stop, self.num, self.endpoint, self.base)

    def _math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            return Logspace(getattr(self.start, operator)(other), getattr(self.stop, operator)(other), self.num, self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))


class Geomspace(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.geomspace>.
    """
    def __init__(self, start, stop, num, endpoint=True, unit=None):
        """
        This is available as a top-level convenience function as <nparray.geomspace>.

        Arguments
        ------------
        * `start` (int or float): the starting point of the sequence.
        * `stop` (int or float): the final value of the sequence, unless `endpoint`
            is False.  In that case, ``num + 1`` values are spaced over the
            interval in log-space, of which all but the last (a sequence of
            length `num`) are returned.
        * `num` (int): number of samples to generate.
        * `endpoint` (bool, optional, default=True): If True, `stop` is the last
            sample. Otherwise, it is not included.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Geomspace>
        """
        super(Geomspace, self).__init__(('start', start, is_float),
                                       ('stop', stop, is_float),
                                       ('num', num, is_int_positive),
                                       ('endpoint', endpoint, is_bool),
                                       ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.geomspace(start, stop, num, endpoint)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.geomspace(self.start, self.stop, self.num, self.endpoint)

    def __math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            # NOTE: we have to cast the start and stop to a float because
            # int.__mul__(float) is not implemented in Python (instead it uses
            # float.__rmul__(int)).
            # Alternatively,we could check for this case and do
            # other.__r{operator}(self.start) but that seems just as ugly.
            return Geomspace(getattr(float(self.start), operator)(other), getattr(float(self.stop), operator)(other), self.num, self.endpoint, self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))

class Full(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.full> or
    <nparray.full_like>.
    """
    def __init__(self, shape, fill_value, unit=None):
        """
        This is available as a top-level convenience function as <nparray.full>
        or <nparray.full_like>.

        Arguments
        ------------
        * `shape` (int or sequence of ints): Shape of the new array, e.g.,
            ``(2, 3)`` or ``2``.
        * `fill_value` (int or float): Value to fill each element in the array.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Full>
        """
        super(Full, self).__init__(('shape', shape, is_valid_shape),
                                    ('fill_value', fill_value, is_float),
                                    ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.full(shape, fill_value)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.full(self.shape, self.fill_value)

    def to_linspace(self):
        """
        Convert from <Full> to <Linspace> (only supported for flat arrays in
        which `shape` is an integer, not a tuple).

        ```py
        Linspace(self.fill_value, self.fill_value, self.shape)
        ```

        Returns
        --------
        * <Linspace>

        Raises
        --------
        * NotImplementedError: if the `shape` is not flat.
        """
        if hasattr(self.shape, '__len__'):
            raise NotImplementedError("can only convert flat Full arrays to linspace")
        return Linspace(self.fill_value, self.fill_value, self.shape)

    def __math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            # NOTE: we have to cast the fill_value to a float because
            # int.__mul__(float) is not implemented in Python (instead it uses
            # float.__rmul__(int)).
            # Alternatively,we could check for this case and do
            # other.__r{operator}(self.fill_value) but that seems just as ugly.
            return Full(self.shape, getattr(float(self.fill_value), operator)(other), self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, Full) or isinstance(other, Zeros) or isinstance(other, Ones):
            if self.shape==other.shape:
                if isinstance(other, Full):
                    other_fill_value = other,fill_value
                elif isinstance(other, Zeros):
                    other_fill_value = 0.0
                elif isinstance(other, Ones):
                    other_fill_value = 1.0
                return Full(self.shape, getattr(self.fill_value, operator)(other_fill_value), self.unit)
            else:
                raise ValueError("cannot add with unmatching shapes")
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))


class Zeros(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.zeros>
    or <nparray.zeros_like>.
    """
    def __init__(self, shape, unit=None):
        """
        This is available as a top-level convenience function as <nparray.zeros>
        or <nparray.zeros_like>.

        Arguments
        ------------
        * `shape` (int or sequence of ints): Shape of the new array, e.g.,
            ``(2, 3)`` or ``2``.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Zeros>
        """
        super(Zeros, self).__init__(('shape', shape, is_valid_shape),
                                    ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.zeros(shape)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.zeros(self.shape)

    def to_full(self):
        """
        Convert from <Zeros> to <Full>.

        ```py
        Full(self.shape, 0)
        ```

        Returns
        ---------
        * <Full>
        """
        return Full(self.shape, 0)

    def to_linspace(self):
        """
        Convert from <Zeros> to <Linspace> (only supported for flat arrays in
        which `shape` is an integer, not a tuple).

        ```py
        Linspace(0, 0, self.shape)
        ```

        Returns
        --------
        * <Linspace>

        Raises
        --------
        * NotImplementedError: if the `shape` is not flat.
        """
        if hasattr(self.shape, '__len__'):
            raise NotImplementedError("can only convert flat Zeros arrays to linspace")
        return Linspace(0, 0, self.shape)

    def __math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            return Full(self.shape, getattr(0.0, operator)(other), self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, Full) or isinstance(other, Zeros) or isinstance(other, Ones):
            if self.shape==other.shape:
                if isinstance(other, Full):
                    other_fill_value = other.fill_value
                elif isinstance(other, Zeros):
                    other_fill_value = 0.0
                elif isinstance(other, Ones):
                    other_fill_value = 1.0
                return Full(self.shape, getattr(0.0, operator)(other_fill_value), self.unit)
            else:
                raise ValueError("cannot add with unmatching shapes")
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))

class Ones(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.ones> or
    <nparray.ones_like>.
    """
    def __init__(self, shape, unit=None):
        """
        This is available as a top-level convenience function as <nparray.ones>
        or <nparray.ones_like>.

        Arguments
        ------------
        * `shape` (int or sequence of ints): Shape of the new array, e.g.,
            ``(2, 3)`` or ``2``.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Ones>
        """
        super(Ones, self).__init__(('shape', shape, is_valid_shape),
                                   ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.ones(shape)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.ones(self.shape)

    def to_full(self):
        """
        Convert from <Ones> to <Full>.

        ```py
        Full(self.shape, 1)
        ```

        Returns
        --------
        * <Full>
        """
        return Full(self.shape, 1)

    def to_linspace(self):
        """
        Convert from <Ones> to <Linspace> (only supported for flat arrays in
        which `shape` is an integer, not a tuple).

        ```py
        Linspace(1, 1, self.shape)
        ```

        Returns
        --------
        * <Linspace>

        Raises
        --------
        * NotImplementedError: if the `shape` is not flat.
        """
        if hasattr(self.shape, '__len__'):
            raise NotImplementedError("can only convert flat Ones arrays to linspace")
        return Linspace(1, 1, self.shape)

    def __math__(self, operator, other):
        if isinstance(other, float) or isinstance(other, int):
            return Full(self.shape, getattr(1.0, operator)(other), self.unit)
        elif isinstance(other, np.ndarray) or isinstance(other, list) or isinstance(other, tuple):
            value = getattr(self.array, operator)(other)
            return Array(value, self.unit)
        elif isinstance(other, Full) or isinstance(other, Zeros) or isinstance(other, Ones):
            if self.shape==other.shape:
                if isinstance(other, Full):
                    other_fill_value = other.fill_value
                elif isinstance(other, Zeros):
                    other_fill_value = 0.0
                elif isinstance(other, Ones):
                    other_fill_value = 1.0
                return Full(self.shape, getattr(1.0, operator)(other_fill_value), self.unit)
            else:
                raise ValueError("cannot add with unmatching shapes")
        elif isinstance(other, ArrayWrapper):
            value = getattr(self.array, operator)(other.array)
            return Array(value, self.unit)
        else:
            raise ValueError("{} not supported with type {}".format(operator, type(other)))

class Eye(ArrayWrapper):
    """
    This is available as a top-level convenience function as <nparray.eye>.
    """
    def __init__(self, M, N=None, k=0, unit=None):
        """
        This is available as a top-level convenience function as <nparray.eye>.

        Arguments
        ------------
        * `M` (int): Number of rows in the output.
        * `N` (int or None, optional, default=None): Number of columns in the output.
            If None, defaults to `N`.
        * `k` (int, optional, default=1): Index of the diagonal: 0 (the default)
            refers to the main diagonal, a positive value refers to an upper
            diagonal, and a negative value to a lower diagonal.
        * `unit` (astropy unit or string, optional, default=None): unit
          corresponding to the passed values.

        Returns
        -----------
        * <Eye>
        """
        super(Eye, self).__init__(('M', M, is_int_positive),
                                  ('N', N, is_int_positive_or_none),
                                  ('k', k, is_int_positive_or_none),
                                  ('unit', unit, is_unit_or_unitstring_or_none))

    @property
    def array(self):
        """
        Compute the underyling numpy array by calling:

        ```py
        np.eye(M, N, k)
        ```

        Returns
        ---------
        * (numpy array): the underlying (computed) numpy array
        """
        return np.eye(self.M, self.N, self.k)

    def __math__(self, operator, other):
        value = getattr(self.to_array(), operator)(other)
        return Array(value, self.unit)



"""
"""

import numpy as np
from scipy.optimize import newton


from phoebe import u, c

try:
    import photodynam
except ImportError:
    _can_bs = False
else:
    _can_bs = True

try:
    import rebound
except ImportError:
    _can_rebound = False
else:
    _can_rebound = True

try:
    import reboundx
except (ImportError, OSError):
    _can_reboundx = False
else:
    _can_reboundx = True

import logging
logger = logging.getLogger("DYNAMICS.NBODY")
logger.addHandler(logging.NullHandler())

_skip_filter_checks = {'check_default': False, 'check_visible': False}

def _ensure_tuple(item):
    """
    Simply ensure that the passed item is a tuple.  If it is not, then
    convert it if possible, or raise a NotImplementedError

    Args:
        item: the item that needs to become a tuple

    Returns:
        the item casted as a tuple

    Raises:
        NotImplementedError: if converting the given item to a tuple
            is not implemented.
    """
    if isinstance(item, tuple):
        return item
    elif isinstance(item, list):
        return tuple(item)
    elif isinstance(item, np.ndarray):
        return tuple(item.tolist())
    else:
        raise NotImplementedError

def dynamics_from_bundle(b, times, compute=None, return_roche_euler=False, use_kepcart=False, **kwargs):
    """
    Parse parameters in the bundle and call :func:`dynamics`.

    See :func:`dynamics` for more detailed information.

    NOTE: you must either provide compute (the label) OR all relevant options
    as kwargs (ltte, stepsize, gr, integrator)

    Args:
        b: (Bundle) the bundle with a set hierarchy
        times: (list or array) times at which to run the dynamics
        stepsize: (float, optional) stepsize for the integration
            [default: 0.01]
        orbiterror: (float, optional) orbiterror for the integration
            [default: 1e-16]
        ltte: (bool, default False) whether to account for light travel time effects.
        gr: (bool, default False) whether to account for general relativity effects.

    Returns:
        t, xs, ys, zs, vxs, vys, vzs.  t is a numpy array of all times,
        the remaining are a list of numpy arrays (a numpy array per
        star - in order given by b.hierarchy.get_stars()) for the cartesian
        positions and velocities of each star at those same times.

    """

    b.run_delayed_constraints()

    hier = b.hierarchy

    computeps = b.get_compute(compute=compute, force_ps=True, **_skip_filter_checks)
    stepsize = computeps.get_value(qualifier='stepsize', stepsize=kwargs.get('stepsize', None), **_skip_filter_checks)
    ltte = computeps.get_value(qualifier='ltte', ltte=kwargs.get('ltte', None), **_skip_filter_checks)
    gr = computeps.get_value(qualifier='gr', gr=kwargs.get('gr', None), **_skip_filter_checks)
    integrator = computeps.get_value(qualifier='integrator', integrator=kwargs.get('integrator', None), **_skip_filter_checks)

    starrefs = hier.get_stars()
    orbitrefs = hier.get_orbits() if use_kepcart else [hier.get_parent_of(star) for star in starrefs]

    def mean_anom(t0, t0_perpass, period):
        # TODO: somehow make this into a constraint where t0 and mean anom
        # are both in the compute options if dynamic_method==nbody
        # (one is constrained from the other and the orbit.... nvm, this gets ugly)
        return 2 * np.pi * (t0 - t0_perpass) / period

    masses = [b.get_value(qualifier='mass', unit=u.solMass, component=component, context='component', **_skip_filter_checks) * c.G.to('AU3 / (Msun d2)').value for component in starrefs]  # GM
    smas = [b.get_value(qualifier='sma', unit=u.AU, component=component, context='component', **_skip_filter_checks) for component in orbitrefs]
    eccs = [b.get_value(qualifier='ecc', component=component, context='component', **_skip_filter_checks) for component in orbitrefs]
    incls = [b.get_value(qualifier='incl', unit=u.rad, component=component, context='component', **_skip_filter_checks) for component in orbitrefs]
    per0s = [b.get_value(qualifier='per0', unit=u.rad, component=component, context='component', **_skip_filter_checks) for component in orbitrefs]
    long_ans = [b.get_value(qualifier='long_an', unit=u.rad, component=component, context='component', **_skip_filter_checks) for component in orbitrefs]
    t0_perpasses = [b.get_value(qualifier='t0_perpass', unit=u.d, component=component, context='component', **_skip_filter_checks) for component in orbitrefs]
    periods = [b.get_value(qualifier='period', unit=u.d, component=component, context='component', **_skip_filter_checks) for component in orbitrefs]

    if return_roche_euler:
        # rotperiods are only needed to compute instantaneous syncpars
        rotperiods = [b.get_value(qualifier='period', unit=u.d, component=component, context='component', **_skip_filter_checks) for component in starrefs]
    else:
        rotperiods = None

    vgamma = b.get_value(qualifier='vgamma', context='system', unit=u.AU/u.d, **_skip_filter_checks)
    t0 = b.get_value(qualifier='t0', context='system', unit=u.d, **_skip_filter_checks)

    # mean_anoms = [mean_anom(t0, t0_perpass, period) for t0_perpass, period in zip(t0_perpasses, periods)]
    mean_anoms = [b.get_value(qualifier='mean_anom', unit=u.rad, component=component, context='component', **_skip_filter_checks) for component in orbitrefs]

    return dynamics(times, masses, smas, eccs, incls, per0s, long_ans, \
                    mean_anoms, rotperiods, t0, vgamma, stepsize, ltte, gr,
                    integrator, use_kepcart=use_kepcart, return_roche_euler=return_roche_euler)


def dynamics(times, masses, smas, eccs, incls, per0s, long_ans, mean_anoms,
        rotperiods=None, t0=0.0, vgamma=0.0, stepsize=0.01, ltte=False, gr=False,
        integrator='ias15', return_roche_euler=False, use_kepcart=False):

    if not _can_rebound:
        raise ImportError("rebound is not installed")

    if gr and not _can_reboundx:
        raise ImportError("reboundx is not installed (required for gr effects)")

    def particle_ltte(sim, particle_N, t_obs):
        c_AU_d = c.c.to(u.AU/u.d).value

        def residual(t):
            # print "*** ltte trying t:", t
            if sim.t != t:
                sim.integrate(t, exact_finish_time=True)
            ltte_dt = sim.particles[particle_N].z / c_AU_d
            t_barycenter = t - ltte_dt
            # print "***** ", t_barycenter-t_obs

            return t_barycenter - t_obs

        t_barycenter = newton(residual, t_obs)

        if sim.t != t_barycenter:
            sim.integrate(t_barycenter)

        return sim.particles[particle_N]


    times = np.asarray(times)

    sim = rebound.Simulation()

    if gr:
        logger.info("enabling 'gr_full' in reboundx")
        rebx = reboundx.Extras(sim)
        # TODO: switch between different GR setups based on masses/hierarchy
        # http://reboundx.readthedocs.io/en/latest/effects.html#general-relativity
        params = rebx.add_gr_full()

    sim.integrator = integrator
    # NOTE: according to rebound docs: "stepsize will change for adaptive integrators such as IAS15"
    sim.dt = stepsize

    if use_kepcart:
        # print "*** bs.kep2cartesian", masses, smas, eccs, incls, per0s, long_ans, mean_anoms, t0
        init_conds = bs.kep2cartesian(_ensure_tuple(masses), _ensure_tuple(smas),
                                      _ensure_tuple(eccs), _ensure_tuple(incls),
                                      _ensure_tuple(per0s), _ensure_tuple(long_ans),
                                      _ensure_tuple(mean_anoms), t0)
        for i in range(len(masses)):
            mass = masses[i]
            x = init_conds['x'][i]
            y = init_conds['y'][i]
            z = init_conds['z'][i]
            vx = init_conds['vx'][i]
            vy = init_conds['vy'][i]
            vz = init_conds['vz'][i]
            # print "*** adding simulation particle for mass:", mass, x, y, z, vx, vy, vz
            sim.add(m=mass,
                    x=x, y=y, z=z,
                    vx=vx, vy=vy, vz=vz
                    )

            # just in case that was stupid and did things in Jacobian, let's for the positions and velocities
            p = sim.particles[-1]
            p.x = x
            p.y = y
            p.z = z
            p.vx = vx
            p.vy = vy
            p.vz = vz
    else:
        for mass, sma, ecc, incl, per0, long_an, mean_anom in zip(masses, smas, eccs, incls, per0s, long_ans, mean_anoms):
            N = sim.N
            # TODO: this assumes building from the inside out (Jacobi coordinates).
            # Make sure that will always happen.
            # This assumption will currently probably fail if inner_as_primary==False.
            # But for now even the inner_as_primary=True case is broken, probably
            # because of the way we define some elements WRT nesting orbits (sma)
            # and others WRT sky (incl, long_an, per0)

            if N==0:
                # print "*** adding simulation particle for mass:", mass
                sim.add(m=mass)
            else:
                # print "*** adding simulation particle for mass:", mass, sma, ecc, incl, long_an, per0, mean_anom
                sim.add(primary=None,
                        m=mass,
                        a=sma,
                        e=ecc,
                        inc=incl,
                        Omega=long_an,
                        omega=per0,
                        M=mean_anom)

            sim.move_to_com()

    # Now handle vgamma by editing the initial vz on each particle
    for particle in sim.particles:
        # vgamma is in the direction of positive RV or negative vz
        particle.vz -= vgamma

    xs = [np.zeros(times.shape) for m in masses]
    ys = [np.zeros(times.shape) for m in masses]
    zs = [np.zeros(times.shape) for m in masses]
    vxs = [np.zeros(times.shape) for m in masses]
    vys = [np.zeros(times.shape) for m in masses]
    vzs = [np.zeros(times.shape) for m in masses]

    if return_roche_euler:
        # from instantaneous Keplerian dynamics for Roche meshing
        ds = [np.zeros(times.shape) for m in masses]
        Fs = [np.zeros(times.shape) for m in masses]

        ethetas = [np.zeros(times.shape) for m in masses]
        elongans = [np.zeros(times.shape) for m in masses]
        eincls = [np.zeros(times.shape) for m in masses]

    au_to_solrad = (1*u.AU).to(u.solRad).value

    for i,time in enumerate(times):

        sim.integrate(time, exact_finish_time=True)

        # if return_roche:
            # TODO: do we need to do this after handling LTTE???
            # orbits = sim.calculate_orbits()

        for j in range(len(masses)):

            if ltte:
                # then we need to integrate to different times per object
                particle = particle_ltte(sim, j, time)
            else:
                # print "***", time, j, sim.N
                particle = sim.particles[j]

            # NOTE: x and y are flipped because of different coordinate system
            # conventions.  If we change our coordinate system to have x point
            # to the left, this will need to be updated to match as well.
            xs[j][i] = -1 * particle.x * au_to_solrad # solRad
            ys[j][i] = -1 * particle.y * au_to_solrad # solRad
            zs[j][i] = particle.z * au_to_solrad  # solRad
            vxs[j][i] = -1 * particle.vx * au_to_solrad # solRad/d
            vys[j][i] = -1 * particle.vy * au_to_solrad # solRad/d
            vzs[j][i] = particle.vz * au_to_solrad # solRad/d

            if return_roche_euler:
                # TODO: do we want the LTTE-adjust particles?

                # NOTE: this won't work for the first particle (as its the
                # primary in the simulation)
                if j==0:
                    particle = sim.particles[j+1]
                else:
                    particle = sim.particles[j]

                # get the orbit based on the primary component defined already
                # in the simulation.
                orbit = particle.calculate_orbit()

                # for instantaneous separation, we need the current separation
                # from the sibling component in units of its instantaneous (?) sma
                ds[j][i] = orbit.d / orbit.a
                # for syncpar (F), assume that the rotational FREQUENCY will
                # remain fixed - so we simply need to updated syncpar based
                # on the INSTANTANEOUS orbital PERIOD.
                Fs[j][i] = orbit.P / rotperiods[j]

                # TODO: need to add np.pi for secondary component
                ethetas[j][i] = orbit.f + orbit.omega # true anomaly + periastron

                elongans[j][i] = orbit.Omega

                eincls[j][i] = orbit.inc


    if return_roche_euler:
        # d, solRad, solRad/d, rad, unitless (sma), unitless, rad, rad, rad
        return times, xs, ys, zs, vxs, vys, vzs, ds, Fs, ethetas, elongans, eincls

    else:
        # d, solRad, solRad/d, rad
        return times, xs, ys, zs, vxs, vys, vzs


def dynamics_from_bundle_bs(b, times, compute=None, return_roche_euler=False, **kwargs):
    """
    Parse parameters in the bundle and call :func:`dynamics`.

    See :func:`dynamics` for more detailed information.

    NOTE: you must either provide compute (the label) OR all relevant options
    as kwargs (ltte)

    Args:
        b: (Bundle) the bundle with a set hierarchy
        times: (list or array) times at which to run the dynamics
        stepsize: (float, optional) stepsize for the integration
            [default: 0.01]
        orbiterror: (float, optional) orbiterror for the integration
            [default: 1e-16]
        ltte: (bool, default False) whether to account for light travel time effects.

    Returns:
        t, xs, ys, zs, vxs, vys, vzs.  t is a numpy array of all times,
        the remaining are a list of numpy arrays (a numpy array per
        star - in order given by b.hierarchy.get_stars()) for the cartesian
        positions and velocities of each star at those same times.

    """

    stepsize = 0.01
    orbiterror = 1e-16
    computeps = b.get_compute(compute, check_visible=False, force_ps=True)
    ltte = computeps.get_value('ltte', check_visible=False, **kwargs)


    hier = b.hierarchy

    starrefs = hier.get_stars()
    orbitrefs = hier.get_orbits()

    def mean_anom(t0, t0_perpass, period):
        # TODO: somehow make this into a constraint where t0 and mean anom
        # are both in the compute options if dynamic_method==nbody
        # (one is constrained from the other and the orbit.... nvm, this gets ugly)
        return 2 * np.pi * (t0 - t0_perpass) / period

    masses = [b.get_value('mass', u.solMass, component=component, context='component') * c.G.to('AU3 / (Msun d2)').value for component in starrefs]  # GM
    smas = [b.get_value('sma', u.AU, component=component, context='component') for component in orbitrefs]
    eccs = [b.get_value('ecc', component=component, context='component') for component in orbitrefs]
    incls = [b.get_value('incl', u.rad, component=component, context='component') for component in orbitrefs]
    per0s = [b.get_value('per0', u.rad, component=component, context='component') for component in orbitrefs]
    long_ans = [b.get_value('long_an', u.rad, component=component, context='component') for component in orbitrefs]
    t0_perpasses = [b.get_value('t0_perpass', u.d, component=component, context='component') for component in orbitrefs]
    periods = [b.get_value('period', u.d, component=component, context='component') for component in orbitrefs]

    vgamma = b.get_value('vgamma', context='system', unit=u.solRad/u.d)
    t0 = b.get_value('t0', context='system', unit=u.d)

    # mean_anoms = [mean_anom(t0, t0_perpass, period) for t0_perpass, period in zip(t0_perpasses, periods)]
    mean_anoms = [b.get_value('mean_anom', u.rad, component=component, context='component') for component in orbitrefs]

    return dynamics_bs(times, masses, smas, eccs, incls, per0s, long_ans, \
                    mean_anoms, t0, vgamma, stepsize, orbiterror, ltte,
                    return_roche_euler=return_roche_euler)




def dynamics_bs(times, masses, smas, eccs, incls, per0s, long_ans, mean_anoms,
        t0=0.0, vgamma=0.0, stepsize=0.01, orbiterror=1e-16, ltte=False,
        return_roche_euler=False):
    """
    Burlisch-Stoer integration of orbits to give positions and velocities
    of any given number of stars in hierarchical orbits.  This code
    currently uses the NBody code in Josh Carter's photodynam code
    available here:

    [[TODO: include link]]

    If using the Nbody mode in PHOEBE, please cite him as well:

    [[TODO: include citation]]

    See :func:`dynamics_from_bundle` for a wrapper around this function
    which automatically handles passing everything in the correct order
    and in the correct units.

    For each iterable input, stars and orbits should be listed in order
    from primary -> secondary for each nested hierarchy level.  Each
    iterable for orbits should have length one less than those for each
    star (ie if 3 masses are provided, then 2 smas, eccs, etc need to
    be provided)

    Args:
        times: (iterable) times at which to compute positions and
            velocities for each star
        masses: (iterable) mass for each star in [solMass]
        smas: (iterable) semi-major axis for each orbit [AU]
        eccs: (iterable) eccentricities for each orbit
        incls: (iterable) inclinations for each orbit [rad]
        per0s: (iterable) longitudes of periastron for each orbit [rad]
        long_ans: (iterable) longitudes of the ascending node for each
            orbit [rad]
        mean_anoms: (iterable) mean anomalies for each orbit
        t0: (float) time at which to start the integrations
        stepsize: (float, optional) stepsize of the integrations
            [default: 0.01]
        orbiterror: (float, optional) orbiterror of the integrations
            [default: 1e-16]
        ltte: (bool, default False) whether to account for light travel time effects.

    Returns:
        t, xs, ys, zs, vxs, vys, vzs.  t is a numpy array of all times,
        the remaining are a list of numpy arrays (a numpy array per
        star - in order given by b.hierarchy.get_stars()) for the cartesian
        positions and velocities of each star at those same times.

    """

    if not _can_bs:
        raise ImportError("photodynam is not installed (http://github.com/phoebe-project/photodynam)")

    times = _ensure_tuple(times)
    masses = _ensure_tuple(masses)
    smas = _ensure_tuple(smas)
    eccs = _ensure_tuple(eccs)
    incls = _ensure_tuple(incls)
    per0s = _ensure_tuple(per0s)
    long_ans = _ensure_tuple(long_ans)
    mean_anoms = _ensure_tuple(mean_anoms)

    # TODO: include vgamma!!!!
    # print "*** bs.do_dynamics", masses, smas, eccs, incls, per0s, long_ans, mean_anoms, t0
    d = photodynam.do_dynamics(times, masses, smas, eccs, incls, per0s, long_ans, mean_anoms, t0, stepsize, orbiterror, ltte, return_roche_euler)
    # d is in the format: {'t': (...), 'x': ( (1,2,3), (1,2,3), ...), 'y': ..., 'z': ...}

    nobjects = len(masses)
    ntimes = len(times)

    # TODO: need to return euler angles... if that even makes sense?? Or maybe we
    # need to make a new place in orbit??

    au_to_solrad = (1*u.AU).to(u.solRad).value

    ts = np.array(d['t'])
    xs = [(-1*np.array([d['x'][ti][oi] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
    ys = [(-1*np.array([d['y'][ti][oi] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
    zs = [(np.array([d['z'][ti][oi] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
    vxs = [(-1*np.array([d['vx'][ti][oi] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
    vys = [(-1*np.array([d['vy'][ti][oi] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
    vzs = [(np.array([d['vz'][ti][oi] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]

    if return_roche_euler:
        # raise NotImplementedError("euler angles for BS not currently supported")
        # a (sma), e (ecc), in (incl), o (per0?), ln (long_an?), m (mean_anom?)
        ds = [(np.array([d['kepl_a'][ti][oi] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
        # TODO: fix this
        Fs = [(np.array([1.0 for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
        # TODO: check to make sure this is the right angle
        # TODO: need to add np.pi for secondary component?
        # true anomaly + periastron
        ethetas = [(np.array([d['kepl_o'][ti][oi]+d['kepl_m'][ti][oi]+np.pi/2 for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
        # elongans = [(np.array([d['kepl_ln'][ti][oi]+long_ans[0 if oi==0 else oi-1] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
        elongans = [(np.array([d['kepl_ln'][ti][oi]+long_ans[0 if oi==0 else oi-1] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
        # eincls = [(np.array([d['kepl_in'][ti][oi]+incls[0 if oi==0 else oi-1] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]
        eincls = [(np.array([d['kepl_in'][ti][oi]+np.pi-incls[0 if oi==0 else oi-1] for ti in range(ntimes)])*au_to_solrad) for oi in range(nobjects)]



        # d, solRad, solRad/d, rad
        return ts, xs, ys, zs, vxs, vys, vzs, ds, Fs, ethetas, elongans, eincls

    else:

        # d, solRad, solRad/d
        return ts, xs, ys, zs, vxs, vys, vzs




import numpy as np

from phoebe import c, u
import libphoebe


def rotfreq_to_omega(rotfreq, M_star, scale=c.R_sun.si.value, solar_units=False):
    """
    Translate from rotation frequency `rotfreq` to `omega`.

    NOTE: everything MUST be in consistent units according to `solar_units`.

    Arguments
    ----------
    * `rotfreq` (float): rotation frequency of the star in cycles/day or cycles/s
        (see `solar_units`)
    * `M_star` (float): **NEW IN 2.3.31** mass of the star (see `solar_units` for units)
    * `scale` (float, optional, default=c.R_sun.si.value)
    * `solar_units` (bool, optional, default=False): whether `rotfreq`, `M_star`,
        and `scale` are provided in solar units (instead of SI).

    Returns
    ---------
    * float
    """
    if solar_units:
        omega = rotfreq / np.sqrt(M_star * c.G.to(u.solRad**3/(u.solMass*u.d**2)).value/scale**3)
    else:
        # then SI units
        omega = rotfreq / np.sqrt(M_star * c.G.value/scale**3)

    return omega

#
# def rpole2potential(rpole, rotfreq, solar_units=False):
#     """
#     Transforms polar radius to surface potential
#     """
#     if not solar_units:
#         rpole = rpole/c.R_sun.si.value
#     rpole_ = np.array([0., 0., rpole])
#     omega = rotfreq_to_omega(rotfreq, solar_units=solar_units)
#     pot =  libphoebe.rotstar_Omega(omega, rpole_)
#     # print "*** rotstar.rpole2potential", rpole, rotfreq, solar_units, omega, pot
#     return pot
#
#
# def potential2rpole(pot, rotfreq, solar_units=False):
#     """
#     Transforms surface potential to polar radius
#     """
#     omega = rotfreq_to_omega(rotfreq, scale=1.0, solar_units=solar_units)
#     # print "*** rotstar.potential2rpole", pot, rotfreq, solar_units, omega
#     rpole = libphoebe.rotstar_pole(omega, pot)
#     if solar_units:
#         return rpole
#     else:
#         return rpole*c.R_sun.si.value




import numpy as np

from phoebe.parameters import *
from phoebe.parameters import dataset as _dataset
import phoebe.dynamics as dynamics
from phoebe.atmospheres import passbands # needed to get choices for 'atm' parameter
from phoebe import u
from phoebe import conf

### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

passbands._init_passbands()  # TODO: move to module import
_atm_choices = list(set([atm for pb in passbands._pbtable.values() for atm in pb['atms'] if atm in passbands._supported_atms]))

def _sampling_params(**kwargs):
    """
    """
    params = []

    params += [SelectParameter(qualifier='sample_from', value=kwargs.get('sample_from', []), choices=[], description='distributions or solutions to use for sampling.  If pointing to a solution, adopt_solution(as_distributions=True, **kwargs) will be called to create a temporary distribution which will be removed automatically.  If all distributions are delta functions (face-values), sample_mode and sample_num will be ignored with a warning.')]
    params += [ChoiceParameter(visible_if='sample_from:<notempty>', qualifier='sample_from_combine', value=kwargs.get('sample_from_combine', 'first'), choices=['first'], description='Method to use to combine multiple distributions from sample_from for the same parameter.  first: ignore duplicate entries and take the first in the sample_from parameter')]
    params += [IntParameter(visible_if='sample_from:<notempty>', qualifier='sample_num', value=kwargs.get('sample_num', 10), limits=(8, 1e6), description='Number of forward models to run sampling from the distributions defined in sample_from and sample_from_combine.')]
    params += [ChoiceParameter(visible_if='sample_from:<notempty>', qualifier='sample_mode', value=kwargs.get('sample_mode', '1-sigma'), choices=['all', 'median', '1-sigma', '3-sigma', '5-sigma'], description='Mode to use when exposing model after sampling.  all: expose all sampled forward-models.  median: only return the median of all sampled models.  1/3/5-sigma: expose the synthetic variable at the median and +/- n-sigma.')]
    params += [BoolParameter(visible_if='sample_from:<notempty>', qualifier='expose_samples', value=kwargs.get('expose_samples', True), description='Whether to expose failed samples along with the simplified error messages.')]
    params += [BoolParameter(visible_if='sample_from:<notempty>', qualifier='expose_failed', value=kwargs.get('expose_failed', True), description='Whether to expose failed samples along with the simplified error messages.')]

    return params

def _comments_params(**kwargs):
    """
    """
    params = []

    params += [StringParameter(qualifier='comments', value=kwargs.get('comments', ''), description='User-provided comments for these compute-options.  Feel free to place any notes here - if not overridden, they will be copied to any resulting models.')]
    return params

def phoebe(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for compute options for the
    PHOEBE 2 backend.  This is the default built-in backend so no other
    pre-requisites are required.

    When using this backend, please see the
    http://phoebe-project.org/publications and cite
    the appropriate references.

    See also:
    * <phoebe.frontend.bundle.Bundle.references>

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_compute>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_compute>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_compute('phoebe')
    b.run_compute(kind='phoebe')
    ```

    Note that default bundles (<phoebe.frontend.bundle.Bundle.default_binary>, for example)
    include a set of compute options for the phoebe backend.

    Arguments
    ----------
    * `enabled` (bool, optional, default=True): whether to create synthetics in
        compute/solver runs.
    * `dynamics_method` (string, optional, default='keplerian'): which method to
        use to determine the dynamics of components.
    * `ltte` (bool, optional, default=False): whether to correct for light
        travel time effects.
    * `atm` (string, optional, default='ck2004'): atmosphere tables.
    * `irrad_method` (string, optional, default='horvat'): which method to use
        to handle irradiation.
    * `boosting_method` (string, optional, default='none'): type of boosting method.
    * `mesh_method` (string, optional, default='marching'): which method to use
        for discretizing the surface.
    * `ntriangles` (int, optional, default=1500): target number of triangles
        (only applicable if `mesh_method` is 'marching').
    * `distortion_method` (string, optional, default='roche'): what type of
        distortion to use when meshing the surface (only applicable
        if `mesh_method` is 'marching').
    * `eclipse_method` (string, optional, default='native'): which method to use
        for determinging eclipses.
    * `lc_method` (string, optional, default='numerical'): which method to use
        for computing light curves.
    * `fti_method` (string, optional, default='oversample'): method to use for
        handling finite-time of integration (exptime).
    * `fti_oversample` (int, optional, default=5): number of times to sample
        per-datapoint for finite-time integration (only applicable if
        `fti_method` is 'oversample').
    * `rv_method` (string, optional, default='flux-weighted'): which method to
        use for computing radial velocities.  If 'dynamical', Rossiter-McLaughlin
        effects will not be computed.
    * `rv_grav` (bool, optional, default=False): whether gravitational redshift
        effects are enabled for RVs (only applicable if `rv_method` is
        'flux-weighted')

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _sampling_params(**kwargs)
    params += _comments_params(**kwargs)

    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'dataset', 'dataset': '*'}, dataset='_default', value=kwargs.get('enabled', True), description='Whether to create synthetics in compute/solver run')]
    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'feature', 'feature': '*'}, feature='_default', value=kwargs.get('enabled', True), description='Whether to enable the feature in compute/solver run')]

    # DYNAMICS
    params += [ChoiceParameter(qualifier='dynamics_method', value=kwargs.get('dynamics_method', 'keplerian'), choices=['keplerian'], description='Which method to use to determine the dynamics of components')]
    params += [BoolParameter(qualifier='ltte', value=kwargs.get('ltte', False), description='Correct for light travel time effects')]

    if conf.devel:
        # note: even though bs isn't an option, its manually added as an option in test_dynamics and test_dynamics_grid
        params += [BoolParameter(visible_if='dynamics_method:bs', qualifier='gr', value=kwargs.get('gr', False), description='Whether to account for general relativity effects')]
        params += [FloatParameter(visible_if='dynamics_method:bs', qualifier='stepsize', value=kwargs.get('stepsize', 0.01), default_unit=None, description='stepsize for the N-body integrator')]         # TODO: improve description (and units??)
        params += [ChoiceParameter(visible_if='dynamics_method:bs', qualifier='integrator', value=kwargs.get('integrator', 'ias15'), choices=['ias15', 'whfast', 'sei', 'leapfrog', 'hermes'], description='Which integrator to use within rebound')]


    # PHYSICS
    # TODO: should either of these be per-dataset... if so: copy_for={'kind': ['rv_dep', 'lc_dep'], 'dataset': '*'}, dataset='_default' and then edit universe.py to pull for the correct dataset (will need to become dataset-dependent dictionary a la ld_func)
    params += [ChoiceParameter(qualifier='irrad_method', value=kwargs.get('irrad_method', 'horvat'), choices=['none', 'wilson', 'horvat'], description='Which method to use to handle all irradiation effects (reflection, redistribution)')]
    params += [ChoiceParameter(qualifier='boosting_method', value=kwargs.get('boosting_method', 'none'), choices=['none'], advanced=True, description='Type of boosting method')]

    # MESH
    # -- these parameters all need to exist per-component --
    # copy_for = {'kind': ['star', 'disk', 'custombody'], 'component': '*'}
    # means that this should exist for each component (since that has a wildcard) which
    # has a kind in [star, disk, custombody]
    # params += [BoolParameter(qualifier='horizon', value=kwargs.get('horizon', False), description='Store horizon for all meshes (except protomeshes)')]
    params += [ChoiceParameter(visible_if='hierarchy.is_meshable:true', copy_for={'kind': ['star', 'envelope'], 'component': '*'},
                               component='_default', qualifier='mesh_method',
                               value=kwargs.get('mesh_method', 'marching'), choices=['marching', 'wd'] if conf.devel else ['marching'],
                               description='Which method to use for discretizing the surface')]

    # NOTE: although the default here is 1500 for ntriangles, add_compute will
    # override this for envelopes already existing in the hierarchy (although
    # any new envelopes in which copy_for triggers a new ntriangles parameter
    # will still get 1500 as a default)
    params += [IntParameter(visible_if='mesh_method:marching,hierarchy.is_meshable:true', copy_for={'kind': ['star', 'envelope'], 'component': '*'}, component='_default', qualifier='ntriangles', value=kwargs.get('ntriangles', 1500), limits=(100,None), default_unit=u.dimensionless_unscaled, description='Requested number of triangles (won\'t be exact).')]
    params += [ChoiceParameter(visible_if='mesh_method:marching,hierarchy.is_meshable:true', copy_for={'kind': ['star'], 'component': '*'}, component='_default', qualifier='distortion_method', value=kwargs.get('distortion_method', 'roche'), choices=['roche', 'rotstar', 'sphere', 'none'], description='Method to use for distorting stars')]

    if conf.devel:
        # TODO: can we have this computed from ntriangles? - and then do the same for the legacy compute options?
        # NOTE: if removing from developer mode - also need to remove if conf.devel in io.py line ~800
        params += [IntParameter(visible_if='mesh_method:wd', copy_for={'kind': ['star', 'envelope'], 'component': '*'}, component='_default', qualifier='gridsize', value=kwargs.get('gridsize', 60), limits=(10,None), default_unit=u.dimensionless_unscaled, description='Number of meshpoints for WD method')]

    if conf.devel:
        params += [BoolParameter(qualifier='mesh_offset', value=kwargs.get('mesh_offset', True), advanced=True, description='Whether to adjust the mesh to have the correct surface area')]
        params += [FloatParameter(visible_if='mesh_method:marching', copy_for={'kind': ['star', 'envelope'], 'component': '*'}, component='_default', qualifier='mesh_init_phi', value=kwargs.get('mesh_init_phi', 0.0), default_unit=u.rad, limits=(0,2*np.pi), advanced=True, description='Initial rotation offset for mesh')]

    # ECLIPSE DETECTION
    params += [ChoiceParameter(qualifier='eclipse_method', value=kwargs.get('eclipse_method', 'native'), choices=['only_horizon', 'graham', 'none', 'visible_partial', 'native', 'wd_horizon'] if conf.devel else ['native', 'only_horizon'], advanced=True, description='Type of eclipse algorithm')]
    params += [ChoiceParameter(visible_if='eclipse_method:native', qualifier='horizon_method', value=kwargs.get('horizon_method', 'boolean'), choices=['boolean', 'linear'] if conf.devel else ['boolean'], advanced=True, description='Type of horizon method')]


    # PER-COMPONENT
    params += [ChoiceParameter(copy_for = {'kind': ['star'], 'component': '*'}, component='_default', qualifier='atm', value=kwargs.get('atm', 'ck2004'), choices=_atm_choices, description='Atmosphere table')]

    # PER-DATASET

    # -- these parameters all need to exist per-rvobs or lcobs --
    # copy_for = {'kind': ['rv_dep'], 'component': '*', 'dataset': '*'}
    # means that this should exist for each component/dataset pair with the
    # rv_dep kind
    # params += [ChoiceParameter(qualifier='lc_method', copy_for = {'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('lc_method', 'numerical'), choices=['numerical', 'analytical'] if conf.devel else ['numerical'], advanced=True, description='Method to use for computing LC fluxes')]
    params += [ChoiceParameter(qualifier='fti_method', copy_for = {'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_method', 'none'), choices=['none', 'oversample'], description='How to handle finite-time integration (when non-zero exptime)')]
    params += [IntParameter(visible_if='fti_method:oversample', qualifier='fti_oversample', copy_for={'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_oversample', 5), limits=(1,None), default_unit=u.dimensionless_unscaled, description='Number of times to sample per-datapoint for finite-time integration')]

    params += [ChoiceParameter(qualifier='rv_method', copy_for={'component': {'kind': 'star'}, 'dataset': {'kind': 'rv'}}, component='_default', dataset='_default', value=kwargs.get('rv_method', 'flux-weighted'), choices=['flux-weighted', 'dynamical'], description='Method to use for computing RVs (must be flux-weighted for Rossiter-McLaughlin effects)')]
    params += [BoolParameter(visible_if='rv_method:flux-weighted', qualifier='rv_grav', copy_for={'component': {'kind': 'star'}, 'dataset': {'kind': 'rv'}}, component='_default', dataset='_default', value=kwargs.get('rv_grav', False), description='Whether gravitational redshift effects are enabled for RVs')]

    if conf.devel:
        params += [ChoiceParameter(qualifier='etv_method', copy_for = {'kind': ['etv'], 'component': '*', 'dataset': '*'}, component='_default', dataset='_default', value=kwargs.get('etv_method', 'crossing'), choices=['crossing'], description='Method to use for computing ETVs')]
        params += [FloatParameter(visible_if='etv_method:crossing', qualifier='etv_tol', copy_for = {'kind': ['etv'], 'component': '*', 'dataset': '*'}, component='_default', dataset='_default', value=kwargs.get('etv_tol', 1e-4), default_unit=u.d, description='Precision with which to determine eclipse timings')]


    return ParameterSet(params)


def legacy(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for compute options for the
    [PHOEBE 1.0 legacy](http://phoebe-project.org/1.0) backend.

    See also:
    * <phoebe.frontend.bundle.Bundle.export_legacy>
    * <phoebe.frontend.bundle.Bundle.from_legacy>

    Use PHOEBE 1.0 (legacy) which is based on the Wilson-Devinney code
    to compute radial velocities and light curves for binary systems
    (>2 stars not supported).  The code is available here:

    http://phoebe-project.org/1.0

    PHOEBE 1.0 and the 'phoebeBackend' python interface must be installed
    and available on the system in order to use this plugin.

    When using this backend, please cite
    * Prsa & Zwitter (2005), ApJ, 628, 426

    See also:
    * <phoebe.frontend.bundle.Bundle.references>

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_compute>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_compute>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_compute('legacy')
    b.run_compute(kind='legacy')
    ```

    Arguments
    ----------
    * `enabled` (bool, optional, default=True): whether to create synthetics in
        compute/solver run.
    * `atm` (string, optional, default='extern_atmx'): atmosphere tables.
    * `pblum_method` (string, optional, default='phoebe'): Method to estimate
        passband luminosities and handle scaling of returned fluxes from legacy.
        stefan-boltzmann: approximate the star as a uniform sphere and estimate
        the luminosities from teff, requiv, logg, and abun from the internal
        passband and atmosphere tables.
        phoebe: build the mesh using roche distortion at time t0 and compute
        luminosities use the internal atmosphere tables (considerable overhead,
        but more accurate for distorted stars).
    * `gridsize` (int, optional, default=60): number of meshpoints for WD.
    * `distortion_method` (string, optional, default='roche'): method to use
        for distorting stars (legacy only supports roche).
    * `irrad_method` (string, optional, default='wilson'): which method to use
        to handle irradiation.
    * `refl_num` (int, optional, default=1): number of reflections (only applicable
        if `irrad_method` is 'wilson').
    * `ie` (bool, optional, default=False): whether data should be de-reddened.
    * `rv_method` (string, optional, default='flux-weighted'): which method to
        use for computing radial velocities.
    * `fti_method` (string, optional, default='none'): How to handle finite-time
        integration (when non-zero exptime)
    * `fit_oversample` (int, optiona, default=5): Number of times to sample
        per-datapoint for finite-time integration (only applicable when `fit_method`
        is 'oversample')

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _sampling_params(**kwargs)
    params += _comments_params(**kwargs)

    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'dataset', 'kind': ['lc', 'rv', 'mesh'], 'dataset': '*'}, dataset='_default', value=kwargs.get('enabled', True), description='Whether to create synthetics in compute/solver run')]
    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'feature', 'kind': ['spot', 'gaussian_process'], 'feature': '*'}, feature='_default', value=kwargs.get('enabled', True), description='Whether to enable the feature in compute/solver run')]

    params += [ChoiceParameter(copy_for = {'kind': ['star'], 'component': '*'}, component='_default', qualifier='atm', value=kwargs.get('atm', 'extern_atmx'), choices=['extern_atmx', 'extern_planckint'], description='Atmosphere table to use within legacy.  For estimating passband luminosities and flux scaling (see pblum_method), extern_atmx will use ck2004 and extern_planckint will use blackbody.')]
    params += [ChoiceParameter(qualifier='pblum_method', value=kwargs.get('pblum_method', 'phoebe'), choices=['stefan-boltzmann', 'phoebe'], description='Method to estimate passband luminosities and handle scaling of returned fluxes from legacy.  stefan-boltzmann: approximate the star as a uniform sphere and estimate the luminosities from teff, requiv, logg, and abun from the internal passband and atmosphere tables.  phoebe: build the mesh using roche distortion at time t0 and compute luminosities use the internal atmosphere tables (considerable overhead, but more accurate for distorted stars).')]

    params += [IntParameter(copy_for={'kind': ['star'], 'component': '*'}, component='_default', qualifier='gridsize', value=kwargs.get('gridsize', 60), limits=(10, 200), description='Number of meshpoints for WD')]

    params += [ChoiceParameter(copy_for={'kind': ['star'], 'component': '*'}, component='_default', qualifier='distortion_method', value=kwargs.get('distortion_method', 'roche'), choices=["roche"], description='Method to use for distorting stars (legacy only supports roche distortion)')]

    params += [ChoiceParameter(qualifier='irrad_method', value=kwargs.get('irrad_method', 'wilson'), choices=['none', 'wilson'], description='Which method to use to handle irradiation/reflection effects')]
    params += [IntParameter(visible_if='irrad_method:wilson', qualifier='refl_num', value=kwargs.get('refl_num', 1), limits=(0,None), description='Number of reflections')]


    # TODO: can we come up with a better qualifier for reddening (and be consistent when we enable in phoebe2)
    params += [BoolParameter(qualifier='ie', value=kwargs.get('ie', False), description='Should data be de-reddened')]

    params += [ChoiceParameter(qualifier='rv_method', copy_for={'component': {'kind': 'star'}, 'dataset': {'kind': 'rv'}}, component='_default', dataset='_default',
                               value=kwargs.get('rv_method', 'flux-weighted'), choices=['flux-weighted', 'dynamical'], description='Method to use for computing RVs (must be flux-weighted for Rossiter-McLaughlin)')]

    params += [ChoiceParameter(qualifier='fti_method', copy_for = {'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_method', 'none'), choices=['none', 'oversample'], description='How to handle finite-time integration (when non-zero exptime)')]
    params += [IntParameter(visible_if='fti_method:oversample', qualifier='fti_oversample', copy_for={'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_oversample', 5), limits=(1,None), default_unit=u.dimensionless_unscaled, description='Number of times to sample per-datapoint for finite-time integration')]


    return ParameterSet(params)

def photodynam(**kwargs):
    """
    **This backend is EXPERIMENTAL and requires developer mode to be enabled**

    **DO NOT USE FOR SCIENCE**

    Create a <phoebe.parameters.ParameterSet> for compute options for Josh
    Carter's [photodynam](http://github.com/phoebe-project/photodynam) code.

    Use photodynam to compute radial velocities and light curves.
    photodynam must be installed and available on the system in order to use
    this plugin.  The code is available here:

    http://github.com/phoebe-project/photodynam

    When using this backend, please cite
    * Science 4 February 2011: Vol. 331 no. 6017 pp. 562-565 DOI:10.1126/science.1201274
    * MNRAS (2012) 420 (2): 1630-1635. doi: 10.1111/j.1365-2966.2011.20151.x

    See also:
    * <phoebe.frontend.bundle.Bundle.references>

    The following parameters are "exported/translated" when using the photodynam
    backend:

    System:
    * t0

    Star:
    * mass
    * requiv

    Orbit:
    * sma
    * ecc
    * incl
    * per0
    * long_an
    * mean_anom

    Dataset:
    * ld_func (only supports quadratic)
    * ld_coeffs (will use <phoebe.frontend.bundle.Bundle.compute_ld_coeffs> if necessary)
    * pblum (will use <phoebe.frontend.bundle.Bundle.compute_pblums> if necessary)


    The following parameters are populated in the resulting model when using the
    photodynam backend:

    LCs:
    * times
    * fluxes

    RVs (dynamical only):
    * times
    * rvs

    ORBs:
    * times
    * us
    * vs
    * ws
    * vus
    * vvs
    * vws

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_compute>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_compute>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_compute('photodynam')
    b.run_compute(kind='photodynam')
    ```

    Arguments
    ----------
    * `enabled` (bool, optional, default=True): whether to create synthetics in
        compute/solver runs.
    * `stepsize` (float, optional, default=0.01): stepsize to use for dynamics
        integration.
    * `orbiterror` (float, optional, default=1e-20): error to use for dynamics
        integration.
    * `distortion_method` (string, optional, default='sphere'): method to use
        for distorting stars (photodynam only supports spherical stars).
    * `irrad_method` (string, optional, default='none'): method to use for
        irradiation (photodynam does not support irradiation).

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    if not conf.devel:
        raise NotImplementedError("'photodynam' backend not officially supported for this release.  Enable developer mode to test.")

    params = _sampling_params(**kwargs)
    params += _comments_params(**kwargs)

    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'dataset', 'kind': ['lc', 'rv', 'orb'], 'dataset': '*'}, dataset='_default', value=kwargs.get('enabled', True), description='Whether to create synthetics in compute/solver run')]
    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'feature', 'kind': ['gaussian_process'], 'feature': '*'}, feature='_default', value=kwargs.get('enabled', True), description='Whether to enable the feature in compute/solver run')]

    params += [ChoiceParameter(copy_for = {'kind': ['star'], 'component': '*'}, component='_default', qualifier='atm', value=kwargs.get('atm', 'ck2004'), advanced=True, choices=_atm_choices, description='Atmosphere table to use when estimating passband luminosities and flux scaling (see pblum_method).  Note photodynam itself does not support atmospheres.')]
    params += [ChoiceParameter(qualifier='pblum_method', value=kwargs.get('pblum_method', 'stefan-boltzmann'), choices=['stefan-boltzmann', 'phoebe'], description='Method to estimate passband luminosities and handle scaling of returned fluxes from photodynam.  stefan-boltzmann: approximate the star as a uniform sphere and estimate the luminosities from teff, requiv, logg, and abun from the internal passband and atmosphere tables.  phoebe: build the mesh using roche distortion at time t0 and compute luminosities use the internal atmosphere tables (considerable overhead, but more accurate for distorted stars).')]

    params += [FloatParameter(qualifier='stepsize', value=kwargs.get('stepsize', 0.01), default_unit=None, description='Stepsize to use for dynamics integration')]
    params += [FloatParameter(qualifier='orbiterror', value=kwargs.get('orbiterror', 1e-20), default_unit=None, description='Error to use for dynamics integraton')]

    params += [ChoiceParameter(copy_for={'kind': ['star'], 'component': '*'}, component='_default', qualifier='distortion_method', value=kwargs.get('distortion_method', 'sphere'), choices=["sphere"], description='Method to use for distorting stars (photodynam only supports spherical stars)')]

    params += [ChoiceParameter(qualifier='irrad_method', value=kwargs.get('irrad_method', 'none'), choices=['none'], description='Which method to use to handle all irradiation effects (ellc does not support irradiation)')]

    params += [ChoiceParameter(qualifier='fti_method', copy_for = {'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_method', 'none'), choices=['none', 'oversample'], description='How to handle finite-time integration (when non-zero exptime)')]
    params += [IntParameter(visible_if='fti_method:oversample', qualifier='fti_oversample', copy_for={'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_oversample', 5), limits=(1,None), default_unit=u.dimensionless_unscaled, description='Number of times to sample per-datapoint for finite-time integration')]


    return ParameterSet(params)

def jktebop(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for compute options for John
    Southworth's [jktebop](https://www.astro.keele.ac.uk/jkt/codes/jktebop.html) code.

    Use jktebop to compute radial velocities and light curves for binary systems.
    jktebop must be installed and available on the system in order to use
    this plugin.  The code is available here (currently tested with v40, requires
    v40+):

    http://www.astro.keele.ac.uk/jkt/codes/jktebop.html

    Please see the link above for a list of publications to cite when using this
    code.

    See also:
    * <phoebe.frontend.bundle.Bundle.references>

    Note on `distortion_method`: according to jktebop's website, "jktebop models
    the two components as biaxial spheroids for the calculation of the reflection
    and ellipsoidal effects, and as spheres for the eclipse shapes."

    Note that the wrapper around jktebop only uses its forward model.
    jktebop also includes its own solver methods, including bootstrapping.
    Those capabilities cannot be accessed from PHOEBE.

    The following parameters are "exported/translated" when using the jktebop
    backend:

    Star:
    * requiv
    * gravb_bol
    * irrad_frac_refl_bol
    * teff (surface-brightness approximated as pblum ratio according to pblum_method divided by square of respective requivs)

    Orbit:
    * sma
    * incl
    * q
    * ecosw
    * esinw
    * period
    * t0_supconj

    Dataset:
    * l3_frac (will be estimated if l3_mode=='flux', but will cost time)
    * ld_mode (cannot be 'interp'.  If 'lookup', coefficients are queried from PHOEBE tables and passed as ld_coeffs)
    * ld_func (supports linear, logarithmic, square_root, quadratic)
    * ld_coeffs (will call <phoebe.frontend.bundle.Bundle.compute_ld_coeffs> if necessary)
    * pblum (will use <phoebe.frontend.bundle.Bundle.compute_pblums> if necessary)

    Note that jktebop works in magnitudes (not fluxes) and is normalized at
    quadrature.  Once converted to fluxes, these are then re-scaled according
    to `pblum_method`.  This renormalization
    is crude and should not be trusted to give absolute fluxes, but should behave
    reasonably with plbum_mode='dataset-scaled'.

    The following parameters are populated in the resulting model when using the
    jktebop backend:

    LCs:
    * times
    * fluxes

    RVs:
    * times
    * rvs

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_compute>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_compute>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_compute('jktebop')
    b.run_compute(kind='jktebop')
    ```

    Arguments
    ----------
    * `enabled` (bool, optional, default=True): whether to create synthetics in
        compute/solver runs.
    * `requiv_max_limit` (float, optional, default=0.5): Maximum allowed fraction
        of requiv_max (as jktebop does not handle highly distorted systems)
        before raising an error in <phoebe.frontend.bundle.Bundle.run_checks_compute>.
    * `atm` (string, optional, default='ck2003'): Atmosphere table to use when
        estimating passband luminosities and flux scaling (see pblum_method).
        Note jktebop itself does not support atmospheres.
    * `pblum_method` (string, optional, default='stefan-boltzmann'): Method to
        estimate passband luminosities and handle scaling of returned fluxes from
        jktebop.  stefan-boltzmann: approximate the star as a uniform sphere and
        estimate the luminosities from teff, requiv, logg, and abun from the
        internal passband and atmosphere tables.  phoebe: build the mesh using
        roche distortion at time t0 and compute luminosities use the internal
         atmosphere tables (considerable overhead, but more accurate for
         distorted stars).
    * `ringsize` (float, optional, default=5): integration ring size.
    * `rv_method` (string, optional, default='dynamical'): Method to use for
        computing RVs.  jktebop only supports dynamical (Keplerian) RVs.
    * `distortion_method` (string, optional, default='sphere/biaxial spheroid'):
        Method to use for distorting stars (applies to all components).
        sphere/biaxial-spheroid: spheres for eclipse shapes and biaxial spheroid
        for calculation of ellipsoidal effects and reflection,
        sphere: sphere for eclipse shapes and no ellipsoidal or reflection effects
    * `irrad_method` (string, optional, default='biaxial spheroid'): method
        to use for computing irradiation.  See note above regarding jktebop's
        treatment of `distortion_method`.
    * `irrad_method` (string, optional, default='none'): method to use for
        irradiation (ellc does not support irradiation).

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _sampling_params(**kwargs)
    params += _comments_params(**kwargs)

    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'dataset', 'kind': ['lc', 'rv'], 'dataset': '*'}, dataset='_default', value=kwargs.get('enabled', True), description='Whether to create synthetics in compute/solver run')]
    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'feature', 'kind': ['gaussian_process'], 'feature': '*'}, feature='_default', value=kwargs.get('enabled', True), description='Whether to enable the feature in compute/solver run')]

    params += [FloatParameter(qualifier='requiv_max_limit', value=kwargs.get('requiv_max_limit', 0.5), limits=(0.01,1), default_unit=u.dimensionless_unscaled, advanced=True, description='Maximum allowed fraction of requiv_max (as jktebop does not handle highly distorted systems) before raising an error in run_checks_compute.')]

    params += [ChoiceParameter(copy_for = {'kind': ['star'], 'component': '*'}, component='_default', qualifier='atm', value=kwargs.get('atm', 'ck2004'), advanced=True, choices=_atm_choices, description='Atmosphere table to use when estimating passband luminosities and flux scaling (see pblum_method).  Note jktebop itself does not support atmospheres.')]
    params += [ChoiceParameter(qualifier='pblum_method', value=kwargs.get('pblum_method', 'stefan-boltzmann'), choices=['stefan-boltzmann', 'phoebe'], description='Method to estimate passband luminosities and handle scaling of returned fluxes from jktebop.  stefan-boltzmann: approximate the star as a uniform sphere and estimate the luminosities from teff, requiv, logg, and abun from the internal passband and atmosphere tables.  phoebe: build the mesh using roche distortion at time t0 and compute luminosities use the internal atmosphere tables (considerable overhead, but more accurate for distorted stars).')]

    params += [FloatParameter(qualifier='ringsize', value=kwargs.get('ringsize', 5), default_unit=u.deg, description='Integration ring size')]

    params += [ChoiceParameter(qualifier='rv_method', copy_for = {'component': {'kind': 'star'}, 'dataset': {'kind': 'rv'}}, component='_default', dataset='_default',
                               value=kwargs.get('rv_method', 'dynamical'), choices=['dynamical'], description='Method to use for computing RVs.  jktebop only supports dynamical (Keplerian) RVs.')]


    params += [ChoiceParameter(qualifier='distortion_method', value=kwargs.get('distortion_method', 'sphere/biaxial spheroid'), choices=["sphere/biaxial spheroid", "sphere"], description='Method to use for distorting stars (applies to all components). sphere/biaxial-spheroid: spheres for eclipse shapes and biaxial spheroid for calculation of ellipsoidal effects and reflection, sphere: sphere for eclipse shapes and no ellipsoidal or reflection effects')]
    params += [ChoiceParameter(qualifier='irrad_method', value=kwargs.get('irrad_method', 'biaxial-spheroid'), choices=['none', 'biaxial-spheroid'], description='Which method to use to handle all irradiation effects')]

    params += [ChoiceParameter(qualifier='fti_method', copy_for = {'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_method', 'none'), choices=['none', 'oversample'], description='How to handle finite-time integration (when non-zero exptime)')]
    params += [IntParameter(visible_if='fti_method:oversample', qualifier='fti_oversample', copy_for={'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_oversample', 5), limits=(1,None), default_unit=u.dimensionless_unscaled, description='Number of times to sample per-datapoint for finite-time integration')]

    return ParameterSet(params)

def ellc(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for compute options for Pierre
    Maxted's [ellc](https://github.com/pmaxted/ellc) code.

    Use ellc to compute radial velocities and light curves for binary systems.
    ellc must be installed and available on the system in order to use
    this plugin (tested with v 1.8.1).  The code is available here:

    https://github.com/pmaxted/ellc

    and can be installed via pip:

    ```py
    pip install ellc
    ```

    Please cite the following when using this backend:

    https://ui.adsabs.harvard.edu/abs/2016A%26A...591A.111M/abstract

    See also:
    * <phoebe.frontend.bundle.Bundle.references>

    Note that the wrapper around ellc only uses its forward model.
    ellc also includes its own solver methods, including emcee.
    Those capabilities cannot be accessed from PHOEBE.

    The following parameters are "exported/translated" when using the ellc
    backend:

    Star:
    * requiv (passed as relative radii by dividing by sma)
    * syncpar
    * gravb_bol
    * teff (surface-brightness approximated as pblum ratio according to pblum_method divided by square of respective requivs)
    * irrad_frac_refl_bol
    * yaw (misalignment only supported with `distortion_method='sphere'` and only included for Rossiter-McLaughlin contribution to RVs)

    Orbit:
    * sma
    * period_anom
    * q
    * incl
    * ecc (passed as `sqrt(ecc)*cos(per0)` and `sqrt(ecc)*sin(per0)`)
    * per0 (passed as `sqrt(ecc)*cos(per0)` and `sqrt(ecc)*sin(per0)`, translated from t0@system to t0_supconj)
    * dperdt (passed as dperdt/period where period is the sidereal period)
    * t0_supconj

    System:
    * vgamma

    Feature (spots only):
    * colat (passed as latitude=-colat)
    * long
    * radius
    * relteff (passed as brightness_factor = relteff^4)

    Dataset (LC/RV only):
    * l3_frac (will be estimated if l3_mode=='flux', but will cost time)
    * ld_mode (cannot be 'interp'.  If 'lookup', coefficients are queried from PHOEBE tables and passed as ld_coeffs)
    * ld_func (supports linear, quadratic, logarithmic, square_root, power)
    * ld_coeffs (will call <phoebe.frontend.bundle.Bundle.compute_ld_coeffs> if necessary)
    * pblum (will use <phoebe.frontend.bundle.Bundle.compute_pblums> if necessary)

    Note: ellc returns fluxes that are normalized based on the sum of the irradiated
    faces of each of the components.  These are then rescaled according to
    `pblum_method`.  Note that this re-normalization is not exact, but should behave
    reasonably with plbum_mode='dataset-scaled'.

    The following parameters are populated in the resulting model when using the
    ellc backend:

    LCs:
    * times
    * fluxes

    RVs:
    * times
    * rvs

    Generally, this will be used as an input to the kind argument in
    <phoebe.frontend.bundle.Bundle.add_compute>.  If attaching through
    <phoebe.frontend.bundle.Bundle.add_compute>, all `**kwargs` will be
    passed on to set the values as described in the arguments below.  Alternatively,
    see <phoebe.parameters.ParameterSet.set_value> to set/change the values
    after creating the Parameters.

    For example:

    ```py
    b.add_compute('ellc')
    b.run_compute(kind='ellc')
    ```

    Arguments
    ----------
    * `enabled` (bool, optional, default=True): whether to create synthetics in
        compute/solver runs.
    * `atm` (string, optional, default='ck2003'): Atmosphere table to use when
        estimating passband luminosities and flux scaling (see pblum_method).
        Note jktebop itself does not support atmospheres.
    * `pblum_method` (string, optional, default='stefan-boltzmann'): Method to
        estimate passband luminosities and handle scaling of returned fluxes from
        jktebop.  stefan-boltzmann: approximate the star as a uniform sphere and
        estimate the luminosities from teff, requiv, logg, and abun from the
        internal passband and atmosphere tables.  phoebe: build the mesh using
        roche distortion at time t0 and compute luminosities use the internal
         atmosphere tables (considerable overhead, but more accurate for
         distorted stars).
    * `distortion_method` (string, optional, default='roche'): method to use
        for distorting stars.
    * `hf` (float, optional, default=1.5): fluid second love number (only applicable
        if/when `distortion_method`='love')
    * `grid` (string, optional, default='default'): grid size used to calculate the flux.
    * `exact_grav` (bool, optional, default=False): whether to use point-by-point
        calculation of local surface gravity for calculation of gravity darkening
        or a (much faster) approximation based on functional form fit to local
        gravity at 4 points on the star.
    * `rv_method` (string, optional, default='flux-weighted'): which method to
        use for computing radial velocities.
    * `irrad_method` (string, optional, default='none'): method to use for
        irradiation (ellc does not support irradiation).
    * `fti_method` (string, optional, default='none'): method to use when accounting
        for finite exposure times.
    * `fti_oversample` (int, optional, default=1): number of integration points
        used to account for finite exposure time.  Only used if `fti_method`='oversample'.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """
    params = _sampling_params(**kwargs)
    params += _comments_params(**kwargs)

    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'dataset', 'kind': ['lc', 'rv'], 'dataset': '*'}, dataset='_default', value=kwargs.get('enabled', True), description='Whether to create synthetics in compute/solver run')]
    params += [BoolParameter(qualifier='enabled', copy_for={'context': 'feature', 'kind': ['spot', 'gaussian_process'], 'feature': '*'}, feature='_default', value=kwargs.get('enabled', True), description='Whether to enable the feature in compute/solver run')]

    params += [ChoiceParameter(copy_for = {'kind': ['star'], 'component': '*'}, component='_default', qualifier='atm', value=kwargs.get('atm', 'ck2004'), advanced=True, choices=_atm_choices, description='Atmosphere table to use when estimating passband luminosities and flux scaling (see pblum_method).  Note ellc itself does not support atmospheres.')]
    params += [ChoiceParameter(qualifier='pblum_method', value=kwargs.get('pblum_method', 'stefan-boltzmann'), choices=['stefan-boltzmann', 'phoebe'], description='Method to estimate passband luminosities and handle scaling of returned fluxes from ellc.  stefan-boltzmann: approximate the star as a uniform sphere and estimate the luminosities from teff, requiv, logg, and abun from the internal passband and atmosphere tables.  phoebe: build the mesh using roche distortion at time t0 and compute luminosities use the internal atmosphere tables (considerable overhead, but more accurate for distorted stars).')]

    params += [ChoiceParameter(copy_for={'kind': ['star'], 'component': '*'}, component='_default', qualifier='distortion_method', value=kwargs.get('distortion_method', 'roche'), choices=["roche", "roche_v", "sphere", "poly1p5", "poly3p0", "love"], description='Method to use for distorting stars')]
    params += [FloatParameter(visible_if='distortion_method:love', copy_for={'kind': ['star'], 'component': '*'}, component='_default', qualifier='hf', value=kwargs.get('hf', 1.5), limits=(0,None), default_unit=u.dimensionless_unscaled, description='fluid second love number for radial displacement')]


    params += [ChoiceParameter(copy_for={'kind': ['star'], 'component': '*'}, component='_default', qualifier='grid', value=kwargs.get('grid', 'default'), choices=['very_sparse', 'sparse', 'default', 'fine', 'very_fine'], description='Grid size used to calculate the flux.')]

    params += [BoolParameter(qualifier='exact_grav', value=kwargs.get('exact_grav', False), description='Whether to use point-by-point calculation of local surface gravity for calculation of gravity darkening or a (much faster) approximation based on functional form fit to local gravity at 4 points on the star.')]

    params += [ChoiceParameter(qualifier='rv_method', copy_for = {'component': {'kind': 'star'}, 'dataset': {'kind': 'rv'}}, component='_default', dataset='_default',
                               value=kwargs.get('rv_method', 'dynamical'), choices=['flux-weighted', 'dynamical'], description='Method to use for computing RVs (must be flux-weighted for Rossiter-McLaughlin).  Note that \'flux-weighted\' is not allowed and will raise an error if irradiation is enabled (see irrad_method).')]


    # copy for RV datasets once exptime support for RVs in phoebe
    params += [ChoiceParameter(qualifier='fti_method', copy_for = {'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_method', 'none'), choices=['none', 'ellc', 'oversample'], description='How to handle finite-time integration (when non-zero exptime).  ellc: use ellcs native oversampling. oversample: use phoebe\'s oversampling')]
    params += [IntParameter(visible_if='fti_method:ellc|oversample', qualifier='fti_oversample', copy_for={'kind': ['lc'], 'dataset': '*'}, dataset='_default', value=kwargs.get('fti_oversample', 5), limits=(1, None), default_unit=u.dimensionless_unscaled, description='number of integration points used to account for finite exposure time.')]

    params += [ChoiceParameter(qualifier='irrad_method', value=kwargs.get('irrad_method', 'lambert'), choices=['lambert', 'none'], description='Which method to use to handle all irradiation effects.  Note that irradiation and rv_method=\'flux-weighted\' cannot be used together.')]

    return ParameterSet(params)

# del deepcopy
# del download_passband, list_installed_passbands, list_online_passbands, list_passbands, parameter_from_json, parse_json, send_if_client, update_if_client
# del fnmatch
